import pkgutil
import importlib


def discover():
    return {
        name: importlib.import_module(name)
        for _, name, _ in pkgutil.iter_modules()
        if name.startswith("turnkeyml_plugin_")
    }



import os
import re
import enum
import sys
import math


class Colors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def log(txt, c=Colors.ENDC, end="", is_error=False):
    logn(txt, c=c, end=end, is_error=is_error)


def logn(txt, c=Colors.ENDC, end="\n", is_error=False):
    file = sys.stderr if is_error else sys.stdout
    print(c + txt + Colors.ENDC, end=end, flush=True, file=file)


class LogType(enum.Enum):
    ERROR = "Error:"
    SUCCESS = "Woohoo!"
    WARNING = "Warning:"
    INFO = "Info:"


def clean_print(type: LogType, msg):
    # Replace path to user’s home directory by a tilde symbol (~)
    home_directory = os.path.expanduser("~")
    home_directory_escaped = re.escape(home_directory)
    msg = re.sub(home_directory_escaped, "~", msg)

    # Split message into list, remove leading spaces and line breaks
    msg = msg.split("\n")
    msg = [line.lstrip() for line in msg]
    while msg[0] == "" and len(msg) > 1:
        msg.pop(0)

    # Print message
    indentation = len(type.value) + 1
    if type == LogType.ERROR:
        log(f"\n{type.value} ".rjust(indentation), c=Colors.FAIL, is_error=True)
    elif type == LogType.SUCCESS:
        log(f"\n{type.value} ".rjust(indentation), c=Colors.OKGREEN)
    elif type == LogType.WARNING:
        log(f"\n{type.value} ".rjust(indentation), c=Colors.WARNING)
    elif type == LogType.INFO:
        log(f"\n{type.value} ".rjust(indentation), c=Colors.OKCYAN)

    is_error = type == LogType.ERROR
    for line_idx, line in enumerate(msg):
        if line_idx != 0:
            log(" " * indentation)
        s_line = line.split("**")
        for idx, l in enumerate(s_line):
            c = Colors.ENDC if idx % 2 == 0 else Colors.BOLD
            if idx != len(s_line) - 1:
                log(l, c=c, is_error=is_error)
            else:
                logn(l, c=c, is_error=is_error)


def log_error(msg):
    clean_print(LogType.ERROR, str(msg))
    # ASCII art credit:
    # https://textart4u.blogspot.com/2014/05/the-fail-whale-ascii-art-code.html
    logn(
        """\n▄██████████████▄▐█▄▄▄▄█▌
██████▌▄▌▄▐▐▌███▌▀▀██▀▀
████▄█▌▄▌▄▐▐▌▀███▄▄█▌
▄▄▄▄▄██████████████\n\n""",
        is_error=True,
    )


def log_success(msg):
    clean_print(LogType.SUCCESS, msg)


def log_warning(msg):
    clean_print(LogType.WARNING, msg)


def log_info(msg):
    clean_print(LogType.INFO, msg)


def list_table(list, padding=25, num_cols=4):
    lines_per_column = int(math.ceil(len(list) / num_cols))
    for i in range(lines_per_column):
        for col in range(num_cols):
            if i + col * lines_per_column < len(list):
                print(
                    list[i + col * lines_per_column].ljust(padding),
                    end="",
                )
        print("\n\t", end="")


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import os
from typing import Optional
import subprocess
from lemonade.state import State
import lemonade.common.status as status
from lemonade.tools import FirstTool
from lemonade.tools.adapter import PassthroughTokenizer, ModelAdapter
from lemonade.cache import Keys
from lemonade.tools.huggingface_load import get_base_model


class LlamaCppAdapter(ModelAdapter):
    def __init__(self, model, output_tokens, context_size, threads, executable):
        super().__init__()

        self.model = os.path.normpath(model)
        self.output_tokens = output_tokens
        self.context_size = context_size
        self.threads = threads
        self.executable = os.path.normpath(executable)

    def generate(
        self,
        input_ids: str,
        max_new_tokens: Optional[int] = None,
        temperature: float = 0.8,
        top_p: float = 0.95,
        top_k: int = 40,
        return_raw: bool = False,
        **kwargs,  # pylint: disable=unused-argument
    ):
        """
        Pass a text prompt into the llamacpp inference CLI.

        The input_ids arg here should receive the original text that
        would normally be encoded by a tokenizer.

        Args:
            input_ids: The input text prompt
            max_new_tokens: Maximum number of tokens to generate
            temperature: Temperature for sampling (0.0 = greedy)
            top_p: Top-p sampling threshold
            top_k: Top-k sampling threshold
            return_raw: If True, returns the complete raw output including timing info
            **kwargs: Additional arguments (ignored)

        Returns:
            List containing a single string with the generated text, or raw output if
            return_raw=True
        """

        prompt = input_ids
        n_predict = max_new_tokens if max_new_tokens is not None else self.output_tokens

        cmd = [
            self.executable,
            "-m",
            self.model,
            "--ctx-size",
            str(self.context_size),
            "-n",
            str(n_predict),
            "-t",
            str(self.threads),
            "-p",
            prompt,
            "--temp",
            str(temperature),
            "--top-p",
            str(top_p),
            "--top-k",
            str(top_k),
            "-e",
            "-no-cnv",
        ]

        cmd = [str(m) for m in cmd]

        try:
            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                encoding="utf-8",
                errors="replace",
            )

            raw_output, stderr = process.communicate(timeout=600)
            if process.returncode != 0:
                error_msg = f"llama.cpp failed with return code {process.returncode}.\n"
                error_msg += f"Command: {' '.join(cmd)}\n"
                error_msg += f"Error output:\n{stderr}\n"
                error_msg += f"Standard output:\n{raw_output}"
                raise Exception(error_msg)

            if raw_output is None:
                raise Exception("No output received from llama.cpp process")

            # Parse timing information
            for line in raw_output.splitlines():
                if "llama_perf_context_print:        eval time =" in line:
                    parts = line.split("(")[1].strip()
                    parts = parts.split(",")
                    ms_per_token = float(parts[0].split("ms per token")[0].strip())
                    self.tokens_per_second = (
                        1000 / ms_per_token if ms_per_token > 0 else 0
                    )
                if "llama_perf_context_print: prompt eval time =" in line:
                    parts = line.split("=")[1].split("/")[0]
                    time_to_first_token_ms = float(parts.split("ms")[0].strip())
                    self.time_to_first_token = time_to_first_token_ms / 1000

            if return_raw:
                return [raw_output, stderr]

            # Find where the prompt ends and the generated text begins
            prompt_found = False
            output_text = ""
            prompt_first_line = prompt.split("\n")[0]
            for line in raw_output.splitlines():
                if prompt_first_line in line:
                    prompt_found = True
                if prompt_found:
                    line = line.replace("</s> [end of text]", "")
                    output_text = output_text + line

            if not prompt_found:
                raise Exception(
                    f"Could not find prompt '{prompt_first_line}' in llama.cpp output. "
                    "This usually means the model failed to process the prompt correctly.\n"
                    f"Raw output:\n{raw_output}\n"
                    f"Stderr:\n{stderr}"
                )

            # Return list containing the generated text
            return [output_text]

        except Exception as e:
            error_msg = f"Failed to run llama.cpp command: {str(e)}\n"
            error_msg += f"Command: {' '.join(cmd)}"
            raise Exception(error_msg)


class LoadLlamaCpp(FirstTool):
    unique_name = "load-llama-cpp"

    def __init__(self):
        super().__init__(monitor_message="Loading llama.cpp model")

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Wrap llama.cpp models with an API",
            add_help=add_help,
        )

        parser.add_argument(
            "--executable",
            required=True,
            type=str,
            help="Path to the llama.cpp executable (e.g., llama-cli or llama-cli.exe)",
        )

        default_threads = 1
        parser.add_argument(
            "--threads",
            required=False,
            type=int,
            default=default_threads,
            help=f"Number of threads to use for generation (default: {default_threads})",
        )

        context_size = 512
        parser.add_argument(
            "--context-size",
            required=False,
            type=int,
            default=context_size,
            help=f"Context size of the prompt (default: {context_size})",
        )

        output_tokens = 512
        parser.add_argument(
            "--output-tokens",
            required=False,
            type=int,
            default=output_tokens,
            help=f"Maximum number of output tokens the LLM should make (default: {output_tokens})",
        )

        parser.add_argument(
            "--model-binary",
            required=True,
            type=str,
            help="Path to a .gguf model file",
        )

        return parser

    def run(
        self,
        state: State,
        input: str = "",
        context_size: int = 512,
        threads: int = 1,
        output_tokens: int = 512,
        model_binary: Optional[str] = None,
        executable: str = None,
    ) -> State:
        """
        Load a llama.cpp model
        """

        if executable is None:
            raise Exception(f"{self.__class__.unique_name} requires an executable path")

        # Convert paths to platform-specific format
        executable = os.path.normpath(executable)

        if model_binary:
            model_to_use = os.path.normpath(model_binary)
        else:
            model_binary = input
            model_to_use = os.path.normpath(model_binary) if model_binary else None

            if not model_binary:
                model_to_use = state.get(Keys.MODEL)

        if model_to_use is None:
            raise Exception(
                f"{self.__class__.unique_name} requires the preceding tool to pass a "
                "Llamacpp model, "
                "or for the user to supply a model with `--model-binary`"
            )

        state.model = LlamaCppAdapter(
            model=model_to_use,
            output_tokens=output_tokens,
            context_size=context_size,
            threads=threads,
            executable=executable,
        )
        state.tokenizer = PassthroughTokenizer()

        # Save stats about the model
        state.save_stat(Keys.CHECKPOINT, model_to_use)

        # Get base model information if this is a converted HF model
        base_model = get_base_model(input)
        if base_model is not None:
            state.save_stat("base_model", base_model)

        status.add_to_state(state=state, name=input, model=model_to_use)

        return state


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import abc


class Profiler(abc.ABC):

    unique_name: str

    def __init__(self, parser_arg_value=None):
        self.parser_arg_value = parser_arg_value
        # Statistics that will be displayed to the CLI user
        self.status_stats = []

    @staticmethod
    @abc.abstractmethod
    def add_arguments_to_parser(parser):
        """
        Adds the argument parsing for this tool to the parser.
        Uses f"--{self.unique_name}" as the argument.
        """

    @abc.abstractmethod
    def start(self, build_dir):
        """
        This method is called prior to the tool sequence starting.
        This informs the profiler to start gathering data.
        The build directory can be used to store profiling data.
        """

    def tool_starting(self, tool_name):
        """
        This method is called to inform the profiler of the name of the tool that is about to start.
        """

    def tool_stopping(self):
        """
        This method is called to inform the profiler that the tool has finished.
        """

    def stop(self):
        """
        This method is called when the tool sequence has finished.
        This informs the profiler to stop gathering data.
        """

    @abc.abstractmethod
    def generate_results(self, state, timestamp, start_times):
        """
        This method is called so that the profiler can create its output files.
        The state is passed so that build info can be gathered and stats can be written.
        The timestamp can be used for filename in current working directory.
        The start times contain a list of tools and start times.
        """


# Copyright (c) 2025 AMD



import argparse
import os
import csv
from typing import Dict, Optional, Any
import requests
from human_eval.data import write_jsonl, read_problems
from human_eval.evaluation import evaluate_functional_correctness

from lemonade.state import State
from lemonade.tools import Tool
import lemonade.common.printing as printing
import lemonade.common.build as build


class AccuracyHumaneval(Tool):
    """
    HumanEval accuracy measurement tool.

    This tool evaluates language models on the HumanEval dataset, which consists of
    Python programming problems. It measures the model's ability to:
    1. Generate functionally correct code completions
    2. Pass unit tests for each programming problem

    Metrics:
    - pass@1: Percentage of problems solved with 1 generation attempt
    - pass@10: Percentage of problems solved within 10 generation attempts
    - pass@100: Percentage of problems solved within 100 generation attempts

    See docs/lemonade/humaneval_accuracy.md for more details
    """

    unique_name = "accuracy-humaneval"
    DATASET = "https://github.com/openai/human-eval/blob/master/data/HumanEval.jsonl.gz?raw=true"
    TOTAL_PROBLEMS = 164  # Total number of problems in the HumanEval dataset

    def __init__(self):
        super().__init__(monitor_message="Measuring accuracy with HumanEval")
        self.status_stats = []
        # Enable code evaluation for HumanEval
        os.environ["HF_ALLOW_CODE_EVAL"] = "1"

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Measure coding accuracy with HumanEval",
            add_help=add_help,
        )
        parser.add_argument(
            "--k-samples",
            type=int,
            default=1,
            help="Number of completions to generate per prompt for pass@k calculation"
            " (default: %(default)s)",
        )
        parser.add_argument(
            "--first-n-samples",
            type=int,
            default=AccuracyHumaneval.TOTAL_PROBLEMS,
            help=f"Evaluate only the first N problems from the dataset (default: "
            f"%(default)s, evaluates all {AccuracyHumaneval.TOTAL_PROBLEMS} problems)",
        )
        parser.add_argument(
            "--timeout",
            type=float,
            default=30.0,
            help="Timeout in seconds for each test case (default: %(default)s)",
        )
        parser.add_argument(
            "--data-dir",
            type=str,
            default=None,
            help="Custom directory for dataset storage (default: %(default)s, "
            "uses <lemonade_cache_dir>/data/humaneval)",
        )
        return parser

    def run(
        self,
        state: State,
        data_dir: Optional[str] = None,
        k_samples: int = 1,
        first_n_samples: Optional[int] = TOTAL_PROBLEMS,
        timeout: float = 30.0,
    ) -> State:
        """
        Run HumanEval evaluation on the model.

        Args:
            state: Current state containing model and tokenizer
            data_dir: Optional custom directory for dataset storage
            k_samples: Number of completions to generate per prompt for pass@k calculation
            first_n_samples: Number of first N problems to evaluate
            timeout: Timeout in seconds for each test case

        Returns:
            Updated state with evaluation results
        """
        # Validate required state components
        if not hasattr(state, "model") or not hasattr(state, "tokenizer"):
            raise ValueError("State must contain both 'model' and 'tokenizer'")

        # Setup directories
        data_dir_to_use = data_dir or os.path.join(state.cache_dir, "data", "humaneval")
        data_path = os.path.join(data_dir_to_use, "HumanEval.jsonl.gz")
        model_results_dir = os.path.join(
            build.output_dir(state.cache_dir, state.build_name), "humaneval"
        )
        os.makedirs(model_results_dir, exist_ok=True)

        # Download dataset if needed
        self._download_dataset(data_path)

        # Run evaluation
        results = self._evaluate_model(
            state.model,
            state.tokenizer,
            data_path,
            k_samples,
            timeout,
            model_results_dir,
            first_n_samples,
        )

        # Save metrics
        self._save_metrics(state, results)

        return state

    def _download_dataset(self, output_path: str) -> None:
        """Download HumanEval dataset if not already present."""
        if os.path.exists(output_path):
            printing.log_info(f"Dataset already exists at: {output_path}")
            return

        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        response = requests.get(self.DATASET, stream=True)

        if response.status_code == 200:
            with open(output_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=8192):
                    file.write(chunk)
            printing.log_info(f"Dataset downloaded successfully to: {output_path}")
        else:
            raise RuntimeError(
                f"Failed to download dataset. Status code: {response.status_code}"
            )

    def _evaluate_model(
        self,
        model: Any,
        tokenizer: Any,
        data_path: str,
        k_samples: int,
        timeout: float,
        results_dir: str,
        first_n_samples: Optional[int] = TOTAL_PROBLEMS,
    ) -> Dict[str, float]:
        """
        Evaluate model on HumanEval dataset.

        Args:
            model: The language model to evaluate
            tokenizer: The tokenizer for the model
            data_path: Path to the HumanEval dataset
            k_samples: Number of completions per prompt for pass@k calculation
            timeout: Test case timeout in seconds
            results_dir: Directory to save results
            first_n_samples: Number of first N problems to evaluate

        Returns:
            Dictionary containing evaluation metrics
        """
        dataset = read_problems(data_path)

        # Limit to first N problems
        dataset_keys = list(dataset.keys())[:first_n_samples]
        ignore_incomplete = True

        samples = []

        # Update Tool progress monitor
        self.set_percent_progress(0.0)
        questions_completed = 0
        number_of_questions = first_n_samples * k_samples

        # Save completions and expected answers
        csv_path = os.path.join(results_dir, "evaluation_results.csv")
        with open(
            csv_path, mode="w", newline="", encoding="utf-8", errors="replace"
        ) as file:
            writer = csv.writer(file)
            writer.writerow(["Prompt", "Completion", "Expected Answer"])

            for task_id in dataset_keys:
                try:
                    for _ in range(k_samples):
                        prompt = dataset[task_id]["prompt"]
                        expected = dataset[task_id]["canonical_solution"]

                        # Generate completion
                        input_ids = tokenizer(prompt, return_tensors="pt").input_ids
                        completion = model.generate(
                            input_ids,
                            max_new_tokens=512,
                            do_sample=False,
                        )
                        completion_text = tokenizer.decode(
                            completion[0], skip_special_tokens=True
                        )

                        # Save results
                        samples.append(
                            {"task_id": task_id, "completion": completion_text}
                        )
                        writer.writerow([prompt, completion_text, expected])

                        # Update progress monitor after completing all samples for a question
                        questions_completed = questions_completed + 1
                        percent_completed = (
                            questions_completed / number_of_questions * 100
                        )
                        self.set_percent_progress(percent_completed)

                # pylint: disable=W0718
                except Exception as e:
                    printing.log_info(f"Error processing task {task_id}: {str(e)}")
                    continue

        # Save predictions and evaluate
        pred_path = os.path.join(results_dir, "humaneval_predictions.jsonl")
        write_jsonl(pred_path, samples)
        printing.log_info(f"Results saved in: {results_dir}")

        # Run functional correctness evaluation
        k_values = [k_samples]
        results = evaluate_functional_correctness(
            pred_path,
            k_values,
            n_workers=1,
            timeout=timeout,
            problem_file=data_path,
            ignore_incomplete=ignore_incomplete,
        )
        return results

    def _save_metrics(self, state: State, results: Dict[str, float]) -> None:
        """Save evaluation metrics to state."""
        for metric, value in results.items():
            metric_name = f"humaneval_{metric}"
            state.save_stat(
                metric_name, float(value) * 100 if value is not None else None
            )
            state.save_stat(f"{metric_name}_units", "%")
            self.status_stats.append(metric_name)


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import os
import tarfile
from pathlib import Path
from typing import List, Optional
import subprocess
import numpy as np
import pandas as pd
import requests
from lemonade.state import State
from lemonade.tools import Tool
import lemonade.common.printing as printing
import lemonade.common.build as build
import lemonade.common.filesystem as fs

# Constants
choices = ["A", "B", "C", "D"]
dataset_url = "https://people.eecs.berkeley.edu/~hendrycks/data.tar"


def min_handle_none(*args: int):
    """
    Returns the minimum of the arguments. If one of the arguments is none,
    it doesn't count towards the min.
    """

    filter_out_none = (value for value in args if value is not None)
    return min(filter_out_none)


class AccuracyMMLU(Tool):
    """
    See docs/lemonade/mmlu_accuracy.md for more details
    """

    unique_name = "accuracy-mmlu"

    def __init__(self):
        super().__init__(monitor_message="Measuring accuracy with MMLU")
        self.status_stats = []

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Measure accuracy with Massive Multitask "
            "Language Understanding (MMLU)",
            add_help=add_help,
        )

        parser.add_argument(
            "--ntrain",
            type=int,
            default=5,
            help="Number of training examples to use. Default set to 5 for `5 Shot`",
        )
        parser.add_argument(
            "--max-evals",
            type=int,
            default=None,
            help="Maximum evaluations to run per test",
        )
        parser.add_argument(
            "--data-dir",
            type=str,
            required=False,
            help="Directory containing test and dev data (default: lemonade cache).",
        )
        parser.add_argument(
            "--tests",
            nargs="+",
            help=(
                "Specific tests to run. For a single quick test, we suggest 'management'."
                + "Default: run all tests."
            ),
        )
        return parser

    def run(
        self,
        state: State,
        ntrain: int = 5,
        max_evals: int = None,
        data_dir: Optional[str] = None,
        tests: List[str] = None,
    ) -> State:

        if data_dir:
            data_dir_to_use = data_dir
        else:
            data_dir_to_use = os.path.join(state.cache_dir, "data", "mmlu")

        # Setup MMLU dataset
        dataset_dir = download_and_extract_dataset(data_dir_to_use, dataset_url)

        model_results_dir = os.path.join(
            build.output_dir(state.cache_dir, state.build_name), "mmlu"
        )
        os.makedirs(model_results_dir, exist_ok=True)

        tests_to_run = [
            f.replace("_test.csv", "")
            for f in sorted(os.listdir(os.path.join(dataset_dir, "test")))
            if f.endswith("_test.csv")
        ]
        if tests is not None:
            unsupported_tests = set(tests) - set(tests_to_run)
            if unsupported_tests:
                raise ValueError(
                    f"Invalid test names provided: {', '.join(unsupported_tests)}. "
                    f"Valid tests are: {', '.join(tests_to_run)}"
                )
            tests_to_run = [test for test in tests if test in tests_to_run]

        tokenizer = state.tokenizer
        model = state.model

        # Update Tool progress monitor
        self.set_percent_progress(0.0)
        number_of_questions = float(
            sum(
                [
                    min_handle_none(
                        len(
                            _safe_read_csv(
                                os.path.join(dataset_dir, "test", f"{subject}_test.csv")
                            )
                        ),
                        max_evals,
                    )
                    for subject in tests_to_run
                ]
            )
        )

        questions_completed = 0

        summary_data = []
        for subject in tests_to_run:
            dev_df = _safe_read_csv(
                os.path.join(dataset_dir, "dev", f"{subject}_dev.csv")
            )[:ntrain]
            test_df = _safe_read_csv(
                os.path.join(dataset_dir, "test", f"{subject}_test.csv")
            )

            # Evaluate the model on the test data for a given subject
            detailed_results = []

            for i in range(min_handle_none(test_df.shape[0], max_evals)):
                prompt = _gen_prompt(dev_df, subject, ntrain) + _format_example(
                    test_df, i, include_answer=False
                )
                input_ids = tokenizer(prompt, return_tensors="pt").input_ids

                response_text = _generate_response(tokenizer, model, input_ids)
                try:
                    pred_label = response_text[-1].upper()
                # Handle models generating empty outputs
                except IndexError:
                    pred_label = "-"

                label = test_df.iloc[i, -1].strip().upper()
                detailed_results.append(
                    {
                        "Question": test_df.iloc[i, 0],
                        "Prompt": prompt,
                        "Correct Answer": label,
                        "Generated Answer": pred_label,
                        "Correct": pred_label == label,
                    }
                )

                # Update progress monitor
                questions_completed = questions_completed + 1
                percent_completed = questions_completed / number_of_questions * 100
                self.set_percent_progress(percent_completed)

            acc = np.mean([res["Correct"] for res in detailed_results])

            subject_results_df = pd.DataFrame(detailed_results)
            subject_csv_path = os.path.join(
                model_results_dir, f"{subject}_detailed_results.csv"
            )
            subject_results_df.to_csv(subject_csv_path, index=False)

            # Update summary_data with total questions and correct answers
            correct_answers_count = sum(
                result["Correct"] for result in detailed_results
            )

            summary_data.append(
                {
                    "Subject": subject,
                    "Accuracy": acc,
                    "Total Questions": len(test_df),
                    "Evaluated Questions": (
                        max_evals
                        if max_evals is not None and max_evals < len(test_df)
                        else len(test_df)
                    ),
                    "Correct Answers": correct_answers_count,
                }
            )

            # Save accuracy results to stats file
            # And display in the CLI
            stat_name = f"mmlu_{subject}_accuracy"
            stat_units_name = f"{stat_name}_units"
            state.save_stat(stat_name, float(acc) * 100)
            state.save_stat(stat_units_name, "%")
            self.status_stats.append(stat_name)

        # Calculate average of mmlu accuracy and display in the CLI
        acc_avg = np.mean([accuracy_data["Accuracy"] for accuracy_data in summary_data])
        state.save_stat(fs.Keys.AVERAGE_MMLU_ACCURACY, float(acc_avg) * 100)
        state.save_stat(f"{fs.Keys.AVERAGE_MMLU_ACCURACY}_units", "%")
        self.status_stats.append(fs.Keys.AVERAGE_MMLU_ACCURACY)

        # Save accuracy results to CSV file
        summary_df = pd.DataFrame(summary_data)
        summary_df.to_csv(
            os.path.join(model_results_dir, "summary_results.csv"), index=False
        )
        return state


def _list_tests(data_dir):
    """Lists all available tests based on the files in the test data directory."""
    test_files = [
        f for f in os.listdir(os.path.join(data_dir, "test")) if f.endswith("_test.csv")
    ]
    print(
        "Available tests:",
        *[f.replace("_test.csv", "") for f in sorted(test_files)],
        sep="\n",
    )


def _format_subject(subject):
    """Formats a subject string by replacing underscores with spaces."""
    return " ".join(subject.split("_"))


def _safe_read_csv(path):
    """Safely reads a CSV file and returns a DataFrame."""
    try:
        return pd.read_csv(path, header=None)
    except FileNotFoundError:
        printing.log_error(f"Error: File not found - {path}")
    except Exception as e:  # pylint: disable=broad-except
        printing.log_error(f"An error occurred while reading {path}: {e}")


def _format_example(df, idx, include_answer=True):
    """Formats an example from the dataframe into a prompt string."""
    prompt = df.iloc[idx, 0]
    for j in range(1, df.shape[1] - 1):
        prompt += f"\n{choices[j-1]}. {df.iloc[idx, j]}"
    prompt += "\nAnswer_:"
    if include_answer:
        prompt += f" {df.iloc[idx, -1]}\n\n"
    return prompt


def _gen_prompt(train_df, subject, k=-1):
    """Generates a prompt string from multiple examples."""
    prompt = (
        "The following are multiple choice questions (with answers) about "
        + f"{_format_subject(subject)}.\n\n"
    )
    for i in range(min(k, train_df.shape[0]) if k != -1 else train_df.shape[0]):
        prompt += _format_example(train_df, i)
    return prompt


def _generate_response(tokenizer, model, input_ids):
    """Generates a model response for the given input IDs."""
    try:
        response = model.generate(input_ids, max_new_tokens=1)
        return tokenizer.decode(response[0], skip_special_tokens=True).strip()
    except subprocess.CalledProcessError as e:
        printing.log_warning(
            f"Subprocess failed with command: {e} and error message: {e.stderr}"
        )
    except Exception as e:  # pylint: disable=broad-except
        printing.log_warning(f"Error during model generation: {e}")
    return ""  # Return an empty string on failure


def download_and_extract_dataset(data_cache_dir: str, dataset_url: str):
    """
    Download the dataset from the given URL and extract it into the target directory.
    """

    # Create the directory if it does not exist
    Path(data_cache_dir).mkdir(parents=True, exist_ok=True)

    # Check if the data already exists to avoid re-downloading
    if not os.listdir(data_cache_dir):  # Checks if the directory is empty
        printing.log_info(f"Downloading dataset to {data_cache_dir}")

        # Download the dataset
        response = requests.get(dataset_url, stream=True)
        if response.status_code == 200:
            tar_path = os.path.join(data_cache_dir, "data.tar")
            with open(tar_path, "wb") as f:
                f.write(response.raw.read())

            printing.log_info("Extracting dataset...")
            # Extract the tar file
            with tarfile.open(tar_path) as tar:
                tar.extractall(path=data_cache_dir)
            os.remove(tar_path)
            printing.log_info("Dataset ready.")
        else:
            printing.log_info("Failed to download the dataset.")
    else:
        printing.log_info(
            f"Dataset already exists in {data_cache_dir}, skipping download."
        )

    # MMLU data is stored in data.tar/data
    return os.path.join(data_cache_dir, "data")


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



from pathlib import Path
import json
from fastapi.responses import HTMLResponse
from lemonade_server.model_manager import ModelManager


def get_instructions_html(port=8000):
    """
    Show instructions on how to use the server.
    """
    # Load server models from JSON
    server_models_path = (
        Path(__file__).parent.parent.parent.parent
        / "lemonade_server"
        / "server_models.json"
    )
    with open(server_models_path, "r", encoding="utf-8") as f:
        server_models = json.load(f)

    # Use shared filter function from model_manager.py
    filtered_models = ModelManager().filter_models_by_backend(server_models)

    # Pass filtered server_models to JS
    server_models_js = (
        f"<script>window.SERVER_MODELS = {json.dumps(filtered_models)};</script>"
    )

    # Load HTML template
    template_path = Path(__file__).parent / "static" / "instructions.html"
    with open(template_path, "r", encoding="utf-8") as f:
        html_template = f.read()

    # Replace template variables
    html_content = html_template.replace("{{SERVER_PORT}}", str(port))
    html_content = html_content.replace("{{SERVER_MODELS_JS}}", server_models_js)

    return HTMLResponse(content=html_content)



"""
Helper functions for dealing with tensors
"""

import os
import copy
import torch
import numpy as np
import lemonade.common.exceptions as exp
import lemonade.common.build as build


# Checks whether a given input has the expected shape
def check_shapes_and_dtypes(
    inputs, expected_shapes, expected_dtypes, expect_downcast=False, raise_error=True
):
    current_shapes, current_dtypes = build.get_shapes_and_dtypes(inputs)

    # If we are modifying the data type of inputs on a later tool we
    # verify input type based on the future data type conversion
    if expect_downcast:
        for key, value in current_dtypes.items():
            if value == "float32":
                current_dtypes[key] = "float16"
            elif value == "int64":
                current_dtypes[key] = "int32"

    input_shapes_changed = expected_shapes != current_shapes
    input_dtypes_changed = expected_dtypes != current_dtypes

    if input_shapes_changed and raise_error:
        msg = f"""
        Model built to always take input of shape
        {expected_shapes} but got {current_shapes}
        """
        raise exp.Error(msg)
    elif input_dtypes_changed and raise_error:
        msg = f"""
        Model built to always take input of types
        {expected_dtypes} but got {current_dtypes}
        """
        raise exp.Error(msg)

    return input_shapes_changed, input_dtypes_changed


def save_inputs(inputs, inputs_file, input_dtypes=None, downcast=True):

    # Detach and downcast inputs
    inputs_converted = copy.deepcopy(inputs)
    for i in range(len(inputs_converted)):
        inputs_converted[i] = {
            k: v for k, v in inputs_converted[i].items() if v is not None
        }
        for k in inputs_converted[i].keys():
            if not hasattr(inputs_converted[i][k], "dtype"):
                continue
            if torch.is_tensor(inputs_converted[i][k]):
                inputs_converted[i][k] = inputs_converted[i][k].cpu().detach().numpy()
            if downcast:
                if input_dtypes is not None and input_dtypes[k] is not None:
                    inputs_converted[i][k] = inputs_converted[i][k].astype(
                        input_dtypes[k]
                    )
                    continue
                if (
                    inputs_converted[i][k].dtype == np.float32
                    or inputs_converted[i][k].dtype == np.float64
                ):
                    inputs_converted[i][k] = inputs_converted[i][k].astype("float16")
                if inputs_converted[i][k].dtype == np.int64:
                    inputs_converted[i][k] = inputs_converted[i][k].astype("int32")

    # Save models inputs to file for later profiling
    if os.path.isfile(inputs_file):
        os.remove(inputs_file)
    np.save(inputs_file, inputs_converted)

    return inputs_converted


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import sys
import time
import os
import platform
import copy
from datetime import datetime
from typing import List, Dict, Optional
import pytz
import psutil
import lemonade.common.printing as printing
import lemonade.common.exceptions as exp
import lemonade.common.build as build
from lemonade.common.system_info import get_system_info_dict
import lemonade.common.filesystem as fs
import lemonade.common.status as status
from lemonade.tools.tool import Tool
from lemonade.profilers.profiler import Profiler
from lemonade.state import State


def _rewind_stdout(lines: int = 1):
    """
    Helper function for the command line monitor. Moves the cursor up a
    certain number of lines in the terminal, corresponding to the
    status line for a Tool, so that we can update the status of
    that Tool.
    """
    rewind_stdout_one_line = "\033[1A"
    rewind_multiple_lines = rewind_stdout_one_line * lines
    print(rewind_multiple_lines, end="")
    sys.stdout.flush()


class Sequence:
    """
    Helper class to launch and manage build tools.
    """

    def __init__(
        self,
        tools: Dict[Tool, List[str]],
        profilers: List[Profiler] = None,
    ):

        self.tools = tools
        self.profilers = [] if profilers is None else profilers

        # Make sure all the tool names are unique
        self.tool_names = [tool.__class__.unique_name for tool in self.tools.keys()]

        if len(self.tool_names) != len(set(self.tool_names)):
            msg = f"""
            All tools in a Sequence must have unique unique_names, however Sequence
            received duplicates in the list of names: {self.tool_names}
            """
            raise ValueError(msg)

        # Save the process (used to get memory usage)
        self.process = psutil.Process()

    def show_monitor(self, state: State, verbosity: bool):
        """
        Displays the monitor on the terminal. The purpose of the monitor
        is to show the status of each tool (success, failure, not started yet,
        or in-progress).
        """

        if verbosity:
            print()

            printing.logn(
                f'Building "{state.build_name}"',
                c=printing.Colors.BOLD,
            )

            for tool in self.tools:
                tool.status_line(successful=None, verbosity=True)

            _rewind_stdout(len(self.tools))

    def _advance_cursor(self, current_tool_name: str):
        # Advance the cursor below the monitor so
        # we can print a message
        tool_depth_in_sequence = len(self.tool_names) - self.tool_names.index(
            current_tool_name
        )
        stdout_lines_to_advance = tool_depth_in_sequence - 2
        cursor_down = "\n" * stdout_lines_to_advance

        print(cursor_down)

    def _get_mem_usage_str(self) -> str:
        """
        Returns a string with memory usage for the current process
        (non-swapped physical memory).  In Windows OS, the peak memory used in the
        process is also included.

        Example: '1.100 GB (1.638 GB peak)'
        """
        mem_info = self.process.memory_info()
        mem_info_str = f"{mem_info.rss / 1024 ** 3:,.3f} GB"
        if platform.system() == "Windows":
            mem_info_str += f" ({mem_info.peak_wset / 1024 ** 3:,.3f} GB peak)"
        return mem_info_str

    def launch(
        self,
        state: State,
        lean_cache: bool = False,
        monitor: Optional[bool] = None,
        stats_to_save: Optional[Dict] = None,
    ) -> State:
        """
        Executes the sequence of tools.
        """

        current_time = datetime.now()
        timestamp = current_time.strftime("%Y-%m-%d-%H%M%S")
        start_times = {"warmup": time.time()}

        # Allow monitor to be globally disabled by an environment variable
        if monitor is None:
            if os.environ.get("LEMONADE_BUILD_MONITOR") == "False":
                monitor_setting = False
            else:
                monitor_setting = True
        else:
            monitor_setting = monitor

        # Create a build directory in the cache
        fs.make_build_dir(state.cache_dir, state.build_name)

        # Start profilers
        build_dir = build.output_dir(state.cache_dir, state.build_name)
        for profiler in self.profilers:
            profiler.start(build_dir)

        self.show_monitor(state, monitor_setting)

        if state.build_status == build.FunctionStatus.SUCCESSFUL:
            msg = """
            build_model() is running a build on a model that already built successfully, which
            should not happen because the build should have loaded from cache or rebuilt from scratch.
            If you are using custom tools and Sequences then you have some debugging to do. Otherwise,
            please file an issue at https://github.com/lemonade-sdk/lemonade/issues
            """
            raise exp.Error(msg)

        # Keep a copy of any stats we loaded from disk, in case we need to
        # restore them later
        saved_stats = copy.deepcopy(fs.Stats(state.cache_dir, state.build_name).stats)

        # Save build name to stats so it shows up on reports
        state.save_stat(fs.Keys.BUILD_NAME, state.build_name)

        # Indicate that the build is running. If the build fails for any reason,
        # we will try to catch the exception and note it in the stats.
        # If a concluded build still has a status of "running", this means
        # there was an uncaught exception.
        state.save_stat(fs.Keys.BUILD_STATUS, build.FunctionStatus.INCOMPLETE)

        # Save a timestamp so that we know the order of builds within a cache
        pacific_tz = pytz.timezone("America/Los_Angeles")
        state.save_stat(
            fs.Keys.TIMESTAMP,
            datetime.now(pacific_tz),
        )

        # Save the system information used for this build
        system_info = get_system_info_dict()
        state.save_stat(
            fs.Keys.SYSTEM_INFO,
            system_info,
        )

        # Collect telemetry for the build
        state.save_stat(
            fs.Keys.SELECTED_SEQUENCE_OF_TOOLS,
            self.tool_names,
        )

        # At the beginning of a sequence no tool has started
        for tool in self.tools:
            state.save_stat(tool.status_key, build.FunctionStatus.NOT_STARTED)
            state.save_stat(tool.duration_key, "-")
            state.save_stat(tool.memory_key, "-")

        # Save any additional stats passed in via arguments
        if stats_to_save:
            for stat_key, stat_value in stats_to_save.items():
                state.save_stat(stat_key, stat_value)

        # Save initial memory as a build statistic
        state.save_stat(f"{fs.Keys.TOOL_MEMORY}:__init__", self._get_mem_usage_str())

        # Run the build
        saved_exception = None
        for tool, argv in self.tools.items():

            start_time = time.time()
            start_times[tool.unique_name] = start_time

            # Inform profiler of name of tool about to start
            for profiler in self.profilers:
                profiler.tool_starting(tool.unique_name)

            try:

                # Set status as incomplete, since tool just started
                state.save_stat(tool.status_key, build.FunctionStatus.INCOMPLETE)

                # Collect telemetry about the tool
                state.current_build_tool = tool.unique_name

                # Run the tool
                state = tool.parse_and_run(state, argv, monitor_setting)

                # Save the state so that it can be assessed for a cache hit
                state.save()

            except exp.SkipBuild as e:
                # SkipBuild is a special exception, which means that a build
                # was loaded from disk, then we realized we want to skip it.
                # In order to preserve the original stats and state of the build,
                # we need to restore the stats file to what it was at the beginning
                # of this function call. We also need to avoid calling state.save().

                # Restore the prior stats
                fs.save_yaml(
                    saved_stats, fs.Stats(state.cache_dir, state.build_name).file
                )

                # Advance the cursor below the monitor so
                # we can print a message
                self._advance_cursor(tool.unique_name)
                printing.log_warning(str(e))
                return

            # Broad exception is desirable as we want to capture
            # all exceptions (including those we can't anticipate)
            except Exception as e:  # pylint: disable=broad-except

                if os.environ.get("LEMONADE_DEBUG", "").lower() == "true":
                    # It may be useful to raise the exception here, since
                    # if any of the subsequent lines of code raise another
                    # exception it will be very hard to root cause e.
                    raise e

                # Update tool and build status
                state.save_stat(tool.status_key, build.FunctionStatus.ERROR)
                state.save_stat(fs.Keys.BUILD_STATUS, build.FunctionStatus.ERROR)

                # Save the log file for the failed tool to stats for easy reference
                stats = fs.Stats(state.cache_dir, state.build_name)
                stats.save_eval_error_log(tool.logfile_path)

                # Advance the cursor below the monitor so
                # we can print a message
                self._advance_cursor(tool.unique_name)

                if vars(state).get("invocation_info"):
                    state.invocation_info.status_message = f"Error: {e}"
                    state.invocation_info.status_message_color = printing.Colors.WARNING
                else:
                    printing.log_error(e)

                # We will raise this exception after we capture as many statistics
                # about the build as possible
                saved_exception = e

                # Don't run any more tools
                break

            else:
                # Update tool Status
                state.save_stat(tool.status_key, build.FunctionStatus.SUCCESSFUL)
                state.current_build_tool = None

            finally:
                # Store tool duration
                execution_time = time.time() - start_time
                state.save_stat(tool.duration_key, execution_time)

                # Store current memory and peak working memory
                state.save_stat(tool.memory_key, self._get_mem_usage_str())

                # Inform profilers that tool has finished
                for profiler in self.profilers:
                    profiler.tool_stopping()

        start_times["cool down"] = time.time()

        # Tell the profilers to stop gathering data
        for profiler in self.profilers:
            profiler.stop()

        if not saved_exception:
            state.build_status = build.FunctionStatus.SUCCESSFUL
            state.save_stat(fs.Keys.BUILD_STATUS, build.FunctionStatus.SUCCESSFUL)
            if vars(state).get("invocation_info"):
                state.invocation_info.status_message = (
                    f"Successful build! {state.invocation_info.extra_status}"
                )
                state.invocation_info.status_message_color = printing.Colors.OKGREEN

        # Generate profiler output
        for profiler in self.profilers:
            profiler.generate_results(state, timestamp, start_times)

        if vars(state).get("models_found") and vars(state).get("invocation_info"):

            # Present status statistics from the tools
            for tool in self.tools:
                state.invocation_info.stats_keys += tool.status_stats

            # Present status statistics from the profilers
            for profiler in self.profilers:
                state.invocation_info.stats_keys += profiler.status_stats

            print()

            status.recursive_print(
                models_found=state.models_found,
                build_name=state.build_name,
                cache_dir=state.cache_dir,
                parent_model_hash=None,
                parent_invocation_hash=None,
                script_names_visited=[],
            )

        if lean_cache:
            printing.log_info("Removing build artifacts...")
            fs.clean_output_dir(state.cache_dir, state.build_name)

        state.save()

        if saved_exception:
            raise saved_exception

        printing.log_success(
            f"\n    Saved to **{build.output_dir(state.cache_dir, state.build_name)}**"
        )

        return state

    def status_line(self, verbosity):
        """
        Print a status line in the monitor for every tool in the sequence
        """
        for tool in self.tools:
            tool.status_line(successful=None, verbosity=verbosity)

    @property
    def info(self) -> Dict[str, Dict]:
        """
        Return a dictionary of tool_name:argv for the sequence
        """

        return {tool.__class__.unique_name: argv for tool, argv in self.tools.items()}


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import statistics
from statistics import StatisticsError
from lemonade.state import State
from lemonade.cache import Keys
from lemonade.tools.adapter import ModelAdapter, TokenizerAdapter
from lemonade.tools.bench import Bench


class OgaBench(Bench):
    """
    Benchmark any model that adheres to the ModelAdapter interface.

    Required input state:
        - MODEL: model instance to benchmark.
        - TOKENIZER: tokenizer instance used to generate inputs for the model.

    Output state produced: None
    """

    unique_name = "oga-bench"

    def __init__(self):
        super().__init__()

        # Additional statistics generated by this bench tool
        self.status_stats.insert(
            self.status_stats.index(Keys.TOKEN_GENERATION_TOKENS_PER_SECOND) + 1,
            Keys.STD_DEV_TOKENS_PER_SECOND,
        )
        self.std_dev_token_generation_tokens_per_second_list = []

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Benchmark an LLM in onnxruntime-genai (OGA)",
            add_help=add_help,
        )

        parser = Bench.parser(parser)

        return parser

    def get_prompt_str(self, state, token_length):
        """
        Returns a string with the prescribed token length.
        """
        tokenizer: TokenizerAdapter = state.tokenizer
        test_prompt = "word " * (token_length - 1)
        input_ids = tokenizer(test_prompt, return_tensors="pt").input_ids
        test_token_length = len(input_ids)
        delta = test_token_length - token_length
        if delta == 0:
            return test_prompt
        return "word " * max(token_length - 1 - delta, 0)

    def run_prompt(
        self,
        state: State,
        report_progress_fn,
        prompt: str,
        iterations: int,
        warmup_iterations: int,
        output_tokens: int,
    ) -> State:

        model: ModelAdapter = state.model
        tokenizer: TokenizerAdapter = state.tokenizer

        input_ids = tokenizer(prompt, return_tensors="pt").input_ids
        self.input_ids_len_list.append(len(input_ids))
        per_iteration_time_to_first_token = []
        per_iteration_tokens_per_second = []

        # Don't capture time for warmup
        for count in range(warmup_iterations):
            outputs = model.generate(input_ids, max_new_tokens=output_tokens)
            self.tokens_out_len_list.append(len(outputs[0]) - len(input_ids))
            report_progress_fn((count + 1) / (warmup_iterations + iterations))

        for count in range(iterations):
            outputs = model.generate(
                input_ids,
                max_new_tokens=output_tokens,
                min_new_tokens=output_tokens,
            )
            report_progress_fn(
                (warmup_iterations + count + 1) / (warmup_iterations + iterations)
            )

            token_len = len(outputs[0]) - len(input_ids)
            self.tokens_out_len_list.append(token_len)

            # Only count an iteration if it produced enough tokens
            if token_len >= output_tokens:
                per_iteration_time_to_first_token.append(model.time_to_first_token)
                per_iteration_tokens_per_second.append(model.tokens_per_second)

        if not per_iteration_time_to_first_token or not per_iteration_tokens_per_second:
            raise Bench.not_enough_tokens(output_tokens)

        mean_time_to_first_token = statistics.mean(per_iteration_time_to_first_token)
        self.mean_time_to_first_token_list.append(mean_time_to_first_token)
        self.prefill_tokens_per_second_list.append(
            len(input_ids) / mean_time_to_first_token
        )
        self.token_generation_tokens_per_second_list.append(
            statistics.mean(per_iteration_tokens_per_second)
        )
        try:
            self.std_dev_time_to_first_token_list.append(
                statistics.stdev(per_iteration_time_to_first_token)
            )
        except StatisticsError:
            # Less than 2 measurements
            self.std_dev_time_to_first_token_list.append(None)
        try:
            self.std_dev_token_generation_tokens_per_second_list.append(
                statistics.stdev(per_iteration_tokens_per_second)
            )
        except StatisticsError:
            # Less than 2 measurements
            self.std_dev_token_generation_tokens_per_second_list.append(None)

    def save_stats(self, state):
        super().save_stats(state)

        # Save additional statistics
        if not all(
            element is None
            for element in self.std_dev_token_generation_tokens_per_second_list
        ):
            state.save_stat(
                Keys.STD_DEV_TOKENS_PER_SECOND,
                self.get_item_or_list(
                    self.std_dev_token_generation_tokens_per_second_list
                ),
            )


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import socketserver
import sys
import logging
from contextlib import asynccontextmanager
from fastapi import FastAPI


def find_free_port():
    """
    Scans for an unoccupied TCP port

    Returns the port number as an int on success
    Returns None if no port can be found
    """

    try:
        with socketserver.TCPServer(("localhost", 0), None) as s:
            return s.server_address[1]
    # pylint: disable=broad-exception-caught
    except Exception:
        return None


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Code here will run when the application starts up
    # Check if console can handle Unicode by testing emoji encoding

    try:
        if sys.stdout.encoding:
            "🍋".encode(sys.stdout.encoding)
        use_emojis = True
    except (UnicodeEncodeError, AttributeError):
        use_emojis = False

    if use_emojis:
        logging.info(
            "\n"
            "\n"
            "🍋  Lemonade Server Ready!\n"
            f"🍋    Open http://localhost:{app.port} in your browser for:\n"
            "🍋      💬 chat\n"
            "🍋      💻 model management\n"
            "🍋      📄 docs\n"
        )
    else:
        logging.info(
            "\n"
            "\n"
            "[Lemonade]  Lemonade Server Ready!\n"
            f"[Lemonade]    Open http://localhost:{app.port} in your browser for:\n"
            "[Lemonade]      chat\n"
            "[Lemonade]      model management\n"
            "[Lemonade]      docs\n"
        )

    yield



import argparse
import json
import os
import socket
import subprocess
import sys
import time
from typing import Optional

import requests

from lemonade.state import State
from lemonade.tools import Tool
import lemonade.common.printing as printing
import lemonade.common.build as build

from lemonade.tools.server.thread_utils import ServerRunner


def is_port_in_use(port, host="localhost"):
    """
    Check if a port is in use
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


class LMEvalHarness(Tool):
    """
    Tool for evaluating LLMs using lm-eval-harness on industry standard benchmarks
    like MMLU, GSM8k, and more. See docs/lemonade/lm_eval.md for more details.
    """

    unique_name = "lm-eval-harness"

    def __init__(self):
        super().__init__(
            monitor_message="Evaluate model accuracy using ElutherAI's lm-eval-harness"
        )
        self.status_stats = []
        self.server_runner = None

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Evaluate model using lm-eval-harness",
            add_help=add_help,
        )

        parser.add_argument(
            "--task",
            type=str,
            required=True,
            help="Task(s) to evaluate on (e.g., gsm8k, mmlu)",
        )

        parser.add_argument(
            "--server-port", type=int, default=8000, help="Port to use for the server"
        )

        parser.add_argument(
            "--num-fewshot",
            type=int,
            default=0,
            help="Number of examples in few-shot prompts",
        )

        parser.add_argument(
            "--limit",
            type=int,
            default=None,
            help="Limit the number of examples per task",
        )

        parser.add_argument(
            "--log-samples",
            action="store_true",
            help="Log samples for each task to log file",
        )

        parser.add_argument(
            "--output-path",
            type=str,
            default=None,
            help="Path to save evaluation results",
        )

        return parser

    def _process_results(self, results_dir, state):
        """Process evaluation results and save to state stats"""
        if not os.path.exists(results_dir) or not os.path.isdir(results_dir):
            printing.log_warning(f"Results directory not found at {results_dir}")
            return

        model_dirs = [
            d
            for d in os.listdir(results_dir)
            if os.path.isdir(os.path.join(results_dir, d))
        ]

        if not model_dirs:
            printing.log_warning(f"No model directories found in {results_dir}")
            return

        model_dir = os.path.join(results_dir, model_dirs[0])
        printing.log_info(f"Found model directory: {model_dir}")

        # Find the results JSON file with timestamp
        results_files = [
            f
            for f in os.listdir(model_dir)
            if f.startswith("results_") and f.endswith(".json")
        ]

        if not results_files:
            printing.log_warning(f"No results files found in {model_dir}")
            return

        # Sort by timestamp
        results_files.sort(reverse=True)
        results_file_path = os.path.join(model_dir, results_files[0])
        printing.log_info(f"Processing results from {results_file_path}")

        # Read and process results
        try:
            with open(results_file_path, "r", encoding="utf-8") as f:
                results = json.load(f)

            # Extract and display metrics
            if "results" in results:
                for task_name, metrics in results["results"].items():
                    printing.log_info(f"Results for {task_name}:")

                    for metric, value in metrics.items():
                        if isinstance(value, (int, float)) and not metric.startswith(
                            "alias"
                        ):
                            # Format metric name for stats
                            clean_metric = metric.replace(",", "_")
                            stat_name = f"lm_eval_{task_name}_{clean_metric}"

                            # Save to state stats as percentage
                            state.save_stat(stat_name, float(value) * 100)
                            state.save_stat(f"{stat_name}_units", "%")
                            self.status_stats.append(stat_name)

                            printing.log_info(
                                f"  {metric}: {value:.4f} ({value*100:.2f}%)"
                            )

                # Save summary metrics if available
                avg_metrics = {}
                if "higher_is_better" in results:
                    for metric_type in results["higher_is_better"].values():
                        for metric in metric_type.keys():
                            if metric not in avg_metrics:
                                avg_metrics[metric] = []

                for task_metrics in results["results"].values():
                    for metric, value in task_metrics.items():
                        if isinstance(value, (int, float)) and not metric.startswith(
                            "alias"
                        ):
                            base_metric = metric.split(",")[0]
                            if base_metric in avg_metrics:
                                avg_metrics[base_metric].append(value)

                # Calculate and save averages
                for metric, values in avg_metrics.items():
                    if values:
                        avg_value = sum(values) / len(values)
                        stat_name = f"lm_eval_average_{metric}"
                        state.save_stat(stat_name, float(avg_value) * 100)
                        state.save_stat(f"{stat_name}_units", "%")
                        self.status_stats.append(stat_name)
                        printing.log_info(
                            f"Average {metric}: {avg_value:.4f} ({avg_value*100:.2f}%)"
                        )

        except (IOError, json.JSONDecodeError) as e:
            printing.log_error(f"Error processing results: {e}")

    def run(
        self,
        state: State,
        task: str,
        server_port: int = 8000,
        server_host: str = "localhost",
        num_fewshot: int = 0,
        limit: Optional[int] = None,
        log_samples: bool = False,
        output_path: Optional[str] = None,
    ) -> State:

        model = state.model
        tokenizer = state.tokenizer

        if model is None or tokenizer is None:
            raise ValueError(
                "Model and tokenizer must be loaded in state before running lm-eval-harness"
            )

        # Set up output path
        if output_path is None:
            output_path = os.path.join(
                build.output_dir(state.cache_dir, state.build_name), "lm_eval_results"
            )

        os.makedirs(output_path, exist_ok=True)

        # Check if port is already in use
        if is_port_in_use(server_port, server_host):
            error_msg = (
                f"Port {server_port} is already in use. "
                "Please close all applications using this port and try again."
            )
            printing.log_error(error_msg)
            raise RuntimeError(error_msg)

        # Retroactively determine recipe based on model type to select correct iterator
        # The model is already loaded in server, so we only need recipe for iterator selection
        checkpoint = getattr(state, "checkpoint", "unknown")
        if "OrtGenaiModel" in str(type(model)):
            recipe = "oga-"
        else:
            recipe = "unknown"

        # Start the server thread
        self.server_runner = ServerRunner(
            model=model,
            tokenizer=tokenizer,
            checkpoint=checkpoint,
            recipe=recipe,
            host=server_host,
            port=server_port,
        )
        self.server_runner.start()

        # Wait for server initialization
        printing.log_info("Waiting for server initialization...")

        # Wait for server to start and be responsive
        server_url = f"http://{server_host}:{server_port}"
        max_retries = 30
        retry_delay = 1

        printing.log_info(f"Checking if server is available at {server_url}...")
        for i in range(max_retries):
            try:
                response = requests.get(f"{server_url}/api/v0/health", timeout=2)
                if response.status_code == 200:
                    printing.log_info(f"Server is ready after {i+1} attempts")
                    break
            except requests.exceptions.RequestException:
                if i < max_retries - 1:
                    time.sleep(retry_delay)
                else:
                    printing.log_error(
                        f"Server did not start after {max_retries} attempts"
                    )
                    raise RuntimeError("Failed to start the server")

        # Build API URL
        results_file = os.path.join(output_path, f"{task}_results")

        printing.log_info(f"Running lm-eval-harness on {task}...")

        # Build lm-eval-harness command
        cmd = [
            "lm_eval",
            "--model",
            "local-completions",
            "--tasks",
            task,
            "--model_args",
            (
                f"model={checkpoint},"
                f"base_url={server_url}/api/v0/completions,"
                f"num_concurrent=1,"
                f"max_retries=5,"
                f"retry_timeout=10,"
                f"tokenized_requests=False"
            ),
            "--num_fewshot",
            str(num_fewshot),
            "--output_path",
            results_file,
        ]

        if limit is not None:
            cmd.extend(["--limit", str(limit)])

        if log_samples:
            cmd.extend(["--log_samples"])

        try:
            # On Windows, set UTF-8 mode to handle Unicode output
            env = os.environ.copy()
            if sys.platform == "win32":
                env["PYTHONIOENCODING"] = "utf-8"

            # Execute lm-eval-harness command
            result = subprocess.run(
                cmd, check=True, text=True, capture_output=True, env=env
            )

            # Log relevant output and skip any parts that might cause encoding issues
            try:
                printing.log_info(result.stdout)
            except UnicodeEncodeError:
                printing.log_info(
                    "Results obtained successfully but couldn't display due to encoding issues"
                )

            # Process results from the correct location
            results_dir = os.path.join(output_path, f"{task}_results")
            self._process_results(results_dir, state)

        except subprocess.CalledProcessError as e:
            printing.log_error(f"Error running lm-eval-harness: {e}")
            printing.log_error(f"stderr: {e.stderr}")
        except (IOError, ValueError, requests.RequestException) as e:
            printing.log_error(f"Error: {e}")
        finally:
            # Shut down server
            if self.server_runner and self.server_runner.is_alive():
                printing.log_info("Shutting down server runner...")
                self.server_runner.shutdown()

            # Make sure we don't have any lingering references to state's model/tokenizer
            # that could prevent garbage collection
            self.server_runner = None

        return state



"""
Helper functions for dealing with ONNX files and ONNX models
"""

import os
from typing import Tuple, Union
import re
import math
import numpy as np
import onnx
import onnxruntime as ort
import lemonade.common.exceptions as exp
from lemonade.state import State
import lemonade.common.build as build


def check_model(onnx_file, success_message, fail_message) -> bool:
    if os.path.isfile(onnx_file):
        print(success_message)
    else:
        print(fail_message)
        return False
    try:
        onnx.checker.check_model(onnx_file)
        print("\tSuccessfully checked onnx file")
        return True
    except onnx.checker.ValidationError as e:
        print("\tError while checking generated ONNX file")
        print(e)
        return False


def original_inputs_file(cache_dir: str, build_name: str):
    return os.path.join(build.output_dir(cache_dir, build_name), "inputs.npy")


def onnx_dir(state: State):
    return os.path.join(build.output_dir(state.cache_dir, state.build_name), "onnx")


def get_output_names(
    onnx_model: Union[str, onnx.ModelProto],
):  # pylint: disable=no-member
    # Get output names of ONNX file/model
    if not isinstance(onnx_model, onnx.ModelProto):  # pylint: disable=no-member
        onnx_model = onnx.load(onnx_model)
    return [node.name for node in onnx_model.graph.output]  # pylint: disable=no-member


def parameter_count(model):
    weights = model.graph.initializer
    parameter_count = 0

    for w in weights:
        weight = onnx.numpy_helper.to_array(w)
        parameter_count += np.prod(weight.shape)
    return parameter_count


def io_bytes(onnx_path: str) -> Tuple[int, int]:
    """Return the number of bytes of each of the inputs and outputs"""
    # pylint: disable = no-member

    def elem_type_to_bytes(elem_type) -> int:
        """
        Convert ONNX's elem_type to the number of bytes used by
        hardware to send that specific datatype through PCIe
        """
        if (
            elem_type == onnx.TensorProto.DataType.UINT8
            or elem_type == onnx.TensorProto.DataType.INT8
            or elem_type == onnx.TensorProto.DataType.BOOL
        ):
            # Each bool requires an entire byte
            return 1
        elif (
            elem_type == onnx.TensorProto.DataType.UINT16
            or elem_type == onnx.TensorProto.DataType.INT16
            or elem_type == onnx.TensorProto.DataType.FLOAT16
        ):
            return 2
        if (
            elem_type == onnx.TensorProto.DataType.FLOAT
            or elem_type == onnx.TensorProto.DataType.INT32
            or elem_type == onnx.TensorProto.DataType.INT64
            or elem_type == onnx.TensorProto.DataType.DOUBLE
            or elem_type == onnx.TensorProto.DataType.UINT64
        ):
            # 64 bit ints are treated as 32 bits everywhere
            # Doubles are treated as floats
            return 4
        elif (
            elem_type == onnx.TensorProto.DataType.COMPLEX64
            or elem_type == onnx.TensorProto.DataType.COMPLEX128
            or elem_type == onnx.TensorProto.DataType.STRING
            or elem_type == onnx.TensorProto.DataType.UNDEFINED
        ):
            raise exp.Error("Unsupported data type")
        else:
            raise exp.Error("Unsupported data type (unknown to ONNX)")

    def get_nodes_bytes(nodes):
        nodes_bytes = {}
        for node in nodes:

            # Get the number of the data type
            dtype_bytes = elem_type_to_bytes(node.type.tensor_type.elem_type)

            # Calculate the total number of elements based on the shape
            shape = str(node.type.tensor_type.shape.dim)
            num_elements = np.prod([int(s) for s in shape.split() if s.isdigit()])

            # Assign a total number of bytes to each node
            nodes_bytes[node.name] = num_elements * dtype_bytes

        return nodes_bytes

    # Get the number of bytes of each of the inputs and outputs
    model = onnx.load(onnx_path)
    onnx_input_bytes = get_nodes_bytes(model.graph.input)
    onnx_output_bytes = get_nodes_bytes(model.graph.output)

    return int(sum(onnx_input_bytes.values())), int(sum(onnx_output_bytes.values()))


def dtype_ort2str(dtype_str: str):
    if dtype_str == "float16":
        datatype = "float16"
    elif dtype_str == "float":
        datatype = "float32"
    elif dtype_str == "double":
        datatype = "float64"
    elif dtype_str == "long":
        datatype = "int64"
    else:
        datatype = dtype_str
    return datatype


def dummy_inputs(onnx_file: str) -> dict:
    # Generate dummy inputs of the expected shape and type for the input model
    sess_options = ort.SessionOptions()
    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
    onnx_session = ort.InferenceSession(onnx_file, sess_options)
    sess_input = onnx_session.get_inputs()

    input_stats = []
    for _idx, input_ in enumerate(range(len(sess_input))):
        input_name = sess_input[input_].name
        input_shape = sess_input[input_].shape

        # TODO: Use onnx update_inputs_outputs_dims to automatically freeze models
        for dim in input_shape:
            if isinstance(dim, str) is True or math.isnan(dim) is True:
                raise AssertionError(
                    "Error: Model has dynamic inputs. Freeze the graph and try again"
                )

        input_type = sess_input[input_].type
        input_stats.append([input_name, input_shape, input_type])

    input_feed = {}
    for stat in input_stats:
        dtype_str = re.search(r"\((.*)\)", stat[2])
        assert dtype_str is not None
        datatype = dtype_ort2str(dtype_str.group(1))
        input_feed[stat[0]] = np.random.rand(*stat[1]).astype(datatype)
    return input_feed


def get_opset(model: onnx.ModelProto) -> int:
    return getattr(model.opset_import[0], "version", None)


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import os
import matplotlib.pyplot as plt
import lemonade.common.build as build
import lemonade.common.printing as printing
from lemonade.state import State
from lemonade.tools import Tool
from lemonade.tools.adapter import ModelAdapter, TokenizerAdapter
from lemonade.cache import Keys

DEFAULT_GENERATE_PARAMS = {
    "do_sample": True,
    "top_k": 50,
    "top_p": 0.95,
    "temperature": 0.7,
}

DEFAULT_RANDOM_SEED = 1
DEFAULT_MAX_NEW_TOKENS = 512
DEFAULT_N_TRIALS = 1


def sanitize_string(input_string):
    return input_string.encode("charmap", "ignore").decode("charmap")


def sanitize_text(text):
    if isinstance(text, str):
        return sanitize_string(text)
    elif isinstance(text, list):
        return [sanitize_string(item) for item in text]
    else:
        raise TypeError("Input must be a string or a list of strings.")


def positive_int(x):
    """Conversion function for argparse"""
    i = int(x)
    if i < 1:
        raise ValueError("Non-positive values are not allowed")
    return i


class LLMPrompt(Tool):
    """
    Send a prompt to an LLM instance and print the response to the screen.

    Required input state:
        - state.model: LLM instance that supports the generate() method.
        - state.tokenizer: LLM tokenizer instance that supports the __call__() (ie, encode)
            and decode() methods.

    Output state produced:
        - "response": text response from the LLM.
    """

    unique_name = "llm-prompt"

    def __init__(self):
        super().__init__(monitor_message="Prompting LLM")

        self.status_stats = [
            Keys.PROMPT_TOKENS,
            Keys.PROMPT,
            Keys.PROMPT_TEMPLATE,
            Keys.RESPONSE_TOKENS,
            Keys.RESPONSE,
            Keys.RESPONSE_LENGTHS_HISTOGRAM,
        ]

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Prompt an LLM and print the result",
            add_help=add_help,
        )

        parser.add_argument(
            "--prompt",
            "-p",
            help="Input prompt to the LLM. Two formats are supported: "
            "1) str: use a user-provided prompt string, and "
            "2) path/to/prompt.txt: load the prompt from a .txt file.",
            required=True,
        )

        parser.add_argument(
            "--template",
            "-t",
            action="store_true",
            help="Insert the prompt into the model's chat template before processing.",
        )

        parser.add_argument(
            "--max-new-tokens",
            "-m",
            default=DEFAULT_MAX_NEW_TOKENS,
            type=int,
            help=f"Maximum number of new tokens in the response "
            f"(default is {DEFAULT_MAX_NEW_TOKENS})",
        )

        parser.add_argument(
            "--n-trials",
            "-n",
            default=DEFAULT_N_TRIALS,
            type=positive_int,
            help=f"Number of responses the LLM will generate for the prompt "
            f"(useful for testing, default is {DEFAULT_N_TRIALS})",
        )

        parser.add_argument(
            "--random-seed",
            "-r",
            default=str(DEFAULT_RANDOM_SEED),
            help="Positive integer seed for random number generator used in "
            "sampling tokens "
            f"(default is {DEFAULT_RANDOM_SEED}). If the number of trials is "
            "greater than one, then the seed is incremented by one for each "
            "trial. Set to `None` for random, non-repeatable results.  This "
            "random seed behavior only applies to models loaded with "
            "`oga-load` or `huggingface-load`.",
        )

        return parser

    def parse(self, state: State, args, known_only=True) -> argparse.Namespace:
        """
        Helper function to parse CLI arguments into the args expected
        by run()
        """

        parsed_args = super().parse(state, args, known_only)

        # Decode prompt arg into a string prompt
        if parsed_args.prompt.endswith(".txt") and os.path.exists(parsed_args.prompt):
            with open(parsed_args.prompt, "r", encoding="utf-8") as f:
                parsed_args.prompt = f.read()

        if parsed_args.random_seed == "None":
            parsed_args.random_seed = None
        else:
            parsed_args.random_seed = int(parsed_args.random_seed)

        return parsed_args

    def run(
        self,
        state: State,
        prompt: str = "Hello",
        max_new_tokens: int = DEFAULT_MAX_NEW_TOKENS,
        n_trials: int = DEFAULT_N_TRIALS,
        template: bool = False,
        random_seed: int = DEFAULT_RANDOM_SEED,
    ) -> State:

        model: ModelAdapter = state.model
        tokenizer: TokenizerAdapter = state.tokenizer

        # If template flag is set, then wrap prompt in template
        if template:
            # Embed prompt in model's chat template
            if tokenizer.chat_template:
                # Use the model's built-in chat template if available
                messages_dict = [{"role": "user", "content": prompt}]
                prompt = tokenizer.apply_chat_template(
                    messages_dict, tokenize=False, add_generation_prompt=True
                )
                state.save_stat(Keys.PROMPT_TEMPLATE, "Model-specific")
            else:
                # Fallback to a standardized template
                printing.log_info("No chat template found. Using default template.")
                prompt = f"<|user|>\n{prompt} <|end|>\n<|assistant|>"
                state.save_stat(Keys.PROMPT_TEMPLATE, "Default")

        input_ids = tokenizer(prompt, return_tensors="pt").input_ids
        if isinstance(input_ids, (list, str)):
            # OGA models return a list of tokens
            # Our llama.cpp adapter returns a string
            len_tokens_in = len(input_ids)
        else:
            # HF models return a 2-D tensor
            len_tokens_in = input_ids.shape[1]

        len_tokens_out = []
        response_texts = []
        for trial in range(n_trials):
            if n_trials > 1:
                self.set_percent_progress(100.0 * trial / n_trials)

            # Get the response from the LLM, which may include the prompt in it
            response = model.generate(
                input_ids,
                max_new_tokens=max_new_tokens,
                random_seed=random_seed,
                **DEFAULT_GENERATE_PARAMS,
            )

            # Increment random seed if not none
            if random_seed is not None:
                random_seed += 1

            # Flatten the input and response
            input_ids_array = (
                input_ids if isinstance(input_ids, (list, str)) else input_ids[0]
            )
            response_array = response if isinstance(response, str) else response[0]

            # Separate the prompt from the response
            len_tokens_out.append(len(response_array) - len_tokens_in)

            input_token = 0
            while (
                input_token < len_tokens_in
                and input_ids_array[input_token] == response_array[input_token]
            ):
                input_token += 1

            # Only decode the actual response (not the prompt)
            response_text = tokenizer.decode(
                response_array[input_token:], skip_special_tokens=True
            ).strip()
            response_texts.append(response_text)

        state.response = response_texts

        if n_trials == 1:
            len_tokens_out = len_tokens_out[0]
            response_texts = response_texts[0]
        else:
            self.set_percent_progress(None)

            # Plot data
            plt.figure()
            plt.hist(len_tokens_out, bins=20)
            plt.xlabel("Response Length (tokens)")
            plt.ylabel("Frequency")
            plt.title(f"Histogram of Response Lengths\n{state.build_name}")
            figure_path = os.path.join(
                build.output_dir(state.cache_dir, state.build_name),
                "response_lengths.png",
            )
            plt.savefig(figure_path)
            state.save_stat(Keys.RESPONSE_LENGTHS_HISTOGRAM, figure_path)

        state.save_stat(Keys.PROMPT_TOKENS, len_tokens_in)
        state.save_stat(Keys.PROMPT, prompt)
        state.save_stat(Keys.RESPONSE_TOKENS, len_tokens_out)
        state.save_stat(Keys.RESPONSE, sanitize_text(response_texts))

        return state


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import json
import os
import huggingface_hub
from importlib.metadata import distributions
from lemonade_server.pydantic_models import LoadConfig


class ModelManager:

    @property
    def supported_models(self) -> dict:
        """
        Returns a dictionary of supported models.
        Note: Models must be downloaded before they are locally available.
        """
        # Load the models dictionary from the JSON file
        server_models_file = os.path.join(
            os.path.dirname(__file__), "server_models.json"
        )
        with open(server_models_file, "r", encoding="utf-8") as file:
            models = json.load(file)

        # Add the model name as a key in each entry, to make it easier
        # to access later

        for key, value in models.items():
            value["model_name"] = key

        return models

    @property
    def downloaded_hf_checkpoints(self) -> list[str]:
        """
        Returns a list of Hugging Face checkpoints that have been downloaded.
        """
        downloaded_hf_checkpoints = []
        try:
            hf_cache_info = huggingface_hub.scan_cache_dir()
            downloaded_hf_checkpoints = [entry.repo_id for entry in hf_cache_info.repos]
        except huggingface_hub.CacheNotFound:
            pass
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"Error scanning Hugging Face cache: {e}")
        return downloaded_hf_checkpoints

    @property
    def downloaded_models(self) -> dict:
        """
        Returns a dictionary of locally available models.
        """
        downloaded_models = {}
        for model in self.supported_models:
            if (
                self.supported_models[model]["checkpoint"].split(":")[0]
                in self.downloaded_hf_checkpoints
            ):
                downloaded_models[model] = self.supported_models[model]
        return downloaded_models

    @property
    def downloaded_models_enabled(self) -> dict:
        """
        Returns a dictionary of locally available models that are enabled by
        the current installation.
        """
        return self.filter_models_by_backend(self.downloaded_models)

    def download_gguf(self, model_config: LoadConfig) -> dict:
        """
        Downloads the GGUF file for the given model configuration.
        """

        # The variant parameter can be either:
        # 1. A full GGUF filename (e.g. "model-Q4_0.gguf")
        # 2. A quantization variant (e.g. "Q4_0")
        # This code handles both cases by constructing the appropriate filename
        checkpoint, variant = model_config.checkpoint.split(":")
        hf_base_name = checkpoint.split("/")[-1].replace("-GGUF", "")
        variant_name = (
            variant if variant.endswith(".gguf") else f"{hf_base_name}-{variant}.gguf"
        )

        # If there is a mmproj file, add it to the patterns
        expected_files = {"variant": variant_name}
        if model_config.mmproj:
            expected_files["mmproj"] = model_config.mmproj

        # Download the files
        snapshot_folder = huggingface_hub.snapshot_download(
            repo_id=checkpoint,
            allow_patterns=list(expected_files.values()),
        )

        # Ensure we downloaded all expected files while creating a dict of the downloaded files
        snapshot_files = {}
        for file in expected_files:
            snapshot_files[file] = os.path.join(snapshot_folder, expected_files[file])
            if expected_files[file] not in os.listdir(snapshot_folder):
                raise ValueError(
                    f"Hugging Face snapshot download for {model_config.checkpoint} "
                    f"expected file {expected_files[file]} not found in {snapshot_folder}"
                )

        # Return a dict that points to the snapshot path of the downloaded GGUF files
        return snapshot_files

    def download_models(self, models: list[str]):
        """
        Downloads the specified models from Hugging Face.
        """
        for model in models:
            if model not in self.supported_models:
                raise ValueError(
                    f"Model {model} is not supported. Please choose from the following: "
                    f"{list(self.supported_models.keys())}"
                )
            checkpoint = self.supported_models[model]["checkpoint"]
            print(f"Downloading {model} ({checkpoint})")

            if "gguf" in checkpoint.lower():
                model_config = LoadConfig(**self.supported_models[model])
                self.download_gguf(model_config)
            else:
                huggingface_hub.snapshot_download(repo_id=checkpoint)

    def filter_models_by_backend(self, models: dict) -> dict:
        """
        Returns a filtered dict of models that are enabled by the
        current environment.
        """
        installed_packages = {dist.metadata["Name"].lower() for dist in distributions()}

        hybrid_installed = (
            "onnxruntime-vitisai" in installed_packages
            and "onnxruntime-genai-directml-ryzenai" in installed_packages
        )
        filtered = {}
        for model, value in models.items():
            if value.get("recipe") == "oga-hybrid":
                if hybrid_installed:
                    filtered[model] = value
            else:
                filtered[model] = value
        return filtered


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



from abc import ABC, abstractmethod
import argparse
import os
import platform
import psutil
from lemonade.state import State
from lemonade.tools import Tool
from lemonade.cache import Keys

default_iterations = 10
default_warmup_runs = 5
default_prompt_length = 64
default_output_tokens = 32
default_prompt = "Hello, I am conscious and"


class Bench(Tool, ABC):
    """
    Abstract parent class for tools that benchmark the performance of the generate()
    method of an LLM.
    """

    def __init__(self, monitor_message="Benchmarking LLM"):
        super().__init__(monitor_message)

        # The minimum set of statistics that a benchmark tool will produce
        # Inherited tools should append any additional statistics they generate to this list
        self.status_stats = [
            Keys.SECONDS_TO_FIRST_TOKEN,
            Keys.STD_DEV_SECONDS_TO_FIRST_TOKEN,
            Keys.TOKEN_GENERATION_TOKENS_PER_SECOND,
            Keys.PREFILL_TOKENS_PER_SECOND,
            Keys.PROMPT_TOKENS,
            Keys.RESPONSE_TOKENS,
            Keys.MAX_MEMORY_USED_GBYTE,
        ]

        # Minimum per measurement statistics
        # Inherited tools should add additional lists for other per prompt statistics
        self.input_ids_len_list = []
        self.tokens_out_len_list = []
        self.mean_time_to_first_token_list = []
        self.std_dev_time_to_first_token_list = []
        self.prefill_tokens_per_second_list = []
        self.token_generation_tokens_per_second_list = []
        self.max_memory_used_gb_list = []

        # Max memory used can only be measured on Windows systems
        self.save_max_memory_used = platform.system() == "Windows"

        # This is set to True only for the duration of the first call to run_prompt
        self.first_run_prompt = None

    @staticmethod
    def parser(parser: argparse.ArgumentParser = None, add_help: bool = True):
        # Allow inherited classes to initialize and pass in a parser, add parameters to it if so
        if parser is None:
            parser = __class__.helpful_parser(
                short_description="Benchmark an LLM", add_help=add_help
            )

        parser.add_argument(
            "--iterations",
            "-i",
            required=False,
            type=int,
            default=default_iterations,
            help="Number of benchmarking iterations to run (default: "
            f"{default_iterations})",
        )

        parser.add_argument(
            "--warmup-iterations",
            "-w",
            required=False,
            type=int,
            default=default_warmup_runs,
            help="Number of benchmarking iterations to use for cache warmup "
            "(the results of these iterations "
            f"are not included in the results; default: {default_warmup_runs})",
        )

        parser.add_argument(
            "--prompts",
            "-p",
            nargs="+",
            required=False,
            default=[str(default_prompt_length)],
            metavar="PROMPT",
            help="Input one or more prompts to the LLM. Three formats are supported. "
            "1) integer: use a synthetic prompt with the specified length "
            "2) str: use a user-provided prompt string "
            "3) path/to/prompt.txt: load the prompt from a text file. "
            f"(default: {default_prompt_length}) ",
        )

        parser.add_argument(
            "--output-tokens",
            required=False,
            type=int,
            default=default_output_tokens,
            help="Number of new tokens the LLM should make (default: "
            f"{default_output_tokens})",
        )

        return parser

    def get_prompt_str(self, _state, token_length):
        """
        Returns a string with approximately the prescribed token length.
        Note: Actual token length is dependent on the tokenizer.
        """
        return "word " * (token_length - 1)

    def parse(self, state: State, args, known_only=True) -> argparse.Namespace:
        """
        Helper function to parse CLI arguments into the args expected by run()
        """

        parsed_args = super().parse(state, args, known_only)

        if parsed_args.prompts is None:
            parsed_args.prompts = [str(default_prompt_length)]

        # Decode prompt arg into a list of prompt strings
        prompt_strings = []
        for prompt_item in parsed_args.prompts:
            if prompt_item.isdigit():
                # Generate a prompt with the requested length
                token_length = int(prompt_item)
                prompt_strings.append(self.get_prompt_str(state, token_length))

            elif os.path.exists(prompt_item):
                with open(prompt_item, "r", encoding="utf-8") as f:
                    prompt_strings.append(f.read())

            else:
                # No change to the prompt
                prompt_strings.append(prompt_item)
        parsed_args.prompts = prompt_strings

        return parsed_args

    def run(
        self,
        state: State,
        prompts: list[str] = None,
        iterations: int = default_iterations,
        warmup_iterations: int = default_warmup_runs,
        output_tokens: int = default_output_tokens,
        **kwargs,
    ) -> State:
        """
        Args:
            - prompts: List of input prompts used as starting points for LLM text generation
            - iterations: number of benchmarking samples to take; results are
                reported as the median and mean of the samples.
            - warmup_iterations: subset of the iterations to treat as warmup,
                and not included in the results.
            - output_tokens: Number of new tokens LLM to create.
            - kwargs: Additional parameters used by bench tools
        """

        if prompts is None:
            prompts = ["word " * (default_prompt_length - 2)]
        elif isinstance(prompts, str):
            prompts = [prompts]

        state.save_stat("prompts", prompts)
        state.save_stat("iterations", iterations)
        state.save_stat("warmup_iterations", warmup_iterations)
        state.save_stat("output_tokens", output_tokens)

        counter = 0
        report_progress_fn = lambda x: self.set_percent_progress(
            100 * (counter + x) / len(prompts)
        )
        self.first_run_prompt = True
        for counter, prompt in enumerate(prompts):
            report_progress_fn(0)

            self.run_prompt(
                state,
                report_progress_fn,
                prompt,
                iterations,
                warmup_iterations,
                output_tokens,
                **kwargs,
            )
            self.first_run_prompt = False

            if self.save_max_memory_used:
                self.max_memory_used_gb_list.append(
                    psutil.Process().memory_info().peak_wset / 1024**3
                )

        self.set_percent_progress(None)
        self.save_stats(state)

        return state

    @abstractmethod
    def run_prompt(
        self,
        state,
        report_progress_fn,
        prompt,
        iterations,
        warmup_iterations,
        output_tokens,
        **kwargs,
    ):
        pass

    @staticmethod
    def get_item_or_list(lst):
        """
        If the list is just a single item then return the item, else return the list
        """
        if len(lst) == 1:
            return lst[0]
        else:
            return lst

    def save_stats(self, state):
        # Save performance data to stats
        state.save_stat(
            Keys.PROMPT_TOKENS, self.get_item_or_list(self.input_ids_len_list)
        )
        state.save_stat(
            Keys.RESPONSE_TOKENS, self.get_item_or_list(self.tokens_out_len_list)
        )
        state.save_stat(
            Keys.SECONDS_TO_FIRST_TOKEN,
            self.get_item_or_list(self.mean_time_to_first_token_list),
        )
        if not all(
            element is None for element in self.std_dev_time_to_first_token_list
        ):
            state.save_stat(
                Keys.STD_DEV_SECONDS_TO_FIRST_TOKEN,
                self.get_item_or_list(self.std_dev_time_to_first_token_list),
            )
        state.save_stat(
            Keys.PREFILL_TOKENS_PER_SECOND,
            self.get_item_or_list(self.prefill_tokens_per_second_list),
        )
        state.save_stat(
            Keys.TOKEN_GENERATION_TOKENS_PER_SECOND,
            self.get_item_or_list(self.token_generation_tokens_per_second_list),
        )
        if self.save_max_memory_used:
            state.save_stat(
                Keys.MAX_MEMORY_USED_GBYTE,
                self.get_item_or_list(self.max_memory_used_gb_list),
            )

    @staticmethod
    def not_enough_tokens(output_tokens: int):
        """
        Raise an exception that explains why a benchmark did not produce any results
        """

        raise ValueError(
            "Your model was benchmarked, however none of the benchmarking "
            "iterations produced the requested amount of output tokens "
            f"(currently {output_tokens}), so "
            "the results have been discarded. You have the following options "
            "to solve this: \n"
            "1. Use the -p option to change the prompt to something that will "
            "produce more output tokens. For example, 'The extremely long "
            "story of my life, told in excruciating details is:' "
            "is an example of a prompt that will result in a lot of output. \n"
            "2. Set a lower value for --output-tokens to make it more likely "
            "that the model will produce enough. \n"
            "3. Set more verbose hyperparameters. \n"
            "4. Run more benchmarking iterations, to improve the chance of "
            "getting at least one with enough output tokens. \n"
        )


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



from setuptools import setup

with open("src/lemonade/version.py", encoding="utf-8") as fp:
    version = fp.read().split('"')[1]

setup(
    name="lemonade-sdk",
    version=version,
    description="Lemonade SDK: Your LLM Aide for Validation and Deployment",
    author_email="lemonade@amd.com",
    package_dir={"": "src"},
    packages=[
        "lemonade",
        "lemonade.profilers",
        "lemonade.common",
        "lemonade.tools",
        "lemonade.tools.ort_genai",
        "lemonade.tools.quark",
        "lemonade.tools.report",
        "lemonade.tools.server",
        "lemonade_install",
        "lemonade_server",
    ],
    install_requires=[
        "invoke>=2.0.0",
        "onnx>=1.11.0,<1.18.0",
        "torch>=1.12.1",
        "pyyaml>=5.4",
        "typeguard>=2.3.13",
        "packaging>=20.9",
        # Necessary until upstream packages account for the breaking
        # change to numpy
        "numpy<2.0.0",
        "pandas>=1.5.3",
        "fasteners",
        "GitPython>=3.1.40",
        "psutil>=6.1.1",
        "wmi",
        "pytz",
        "zstandard",
        "matplotlib",
        "tabulate",
        # huggingface-hub==0.31.0 introduces a new transfer protocol that was causing us issues
        "huggingface-hub==0.30.2",
    ],
    extras_require={
        "llm": [
            "torch>=2.6.0",
            "transformers<=4.51.3",
            "accelerate",
            "py-cpuinfo",
            "sentencepiece",
            "datasets",
            # Install human-eval from a forked repo with Windows support until the
            # PR (https://github.com/openai/human-eval/pull/53) is merged
            "human-eval-windows==1.0.4",
            "fastapi",
            "uvicorn[standard]",
            "openai>=1.81.0",
            "lm-eval[api]",
        ],
        "llm-oga-cpu": [
            "onnxruntime-genai==0.6.0",
            "onnxruntime >=1.10.1,<1.22.0",
            "lemonade-sdk[llm]",
        ],
        "llm-oga-igpu": [
            "onnxruntime-genai-directml==0.6.0",
            "onnxruntime-directml>=1.19.0,<1.22.0",
            "transformers<4.45.0",
            "lemonade-sdk[llm]",
        ],
        "llm-oga-cuda": [
            "onnxruntime-genai-cuda==0.6.0",
            "onnxruntime-gpu >=1.19.1,<1.22.0",
            "transformers<4.45.0",
            "lemonade-sdk[llm]",
        ],
        "llm-oga-npu": [
            "onnx==1.16.0",
            "onnxruntime==1.18.0",
            "numpy==1.26.4",
            "protobuf>=6.30.1",
            "lemonade-sdk[llm]",
        ],
        "llm-oga-hybrid": [
            "onnx==1.16.1",
            "numpy==1.26.4",
            "protobuf>=6.30.1",
            "lemonade-sdk[llm]",
        ],
        "llm-oga-unified": [
            "lemonade-sdk[llm-oga-hybrid]",
        ],
    },
    classifiers=[],
    entry_points={
        "console_scripts": [
            "lemonade=lemonade:lemonadecli",
            "lemonade-install=lemonade_install:installcli",
            "lemonade-server-dev=lemonade_server.cli:main",
        ]
    },
    python_requires=">=3.10, <3.12",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    package_data={
        "lemonade_server": ["server_models.json"],
        "lemonade": ["tools/server/static/*"],
    },
)

# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



# onnxruntime_genai is not lint-friendly yet and PyLint can't
# find any of the class methods
# pylint: disable=no-member
#
# Model builder constraints:
#   11/10/24 Need transformers <4.45.0 OR onnxruntime-genai 0.5.0 (which must be built from source)
#   (transformers v4.45 changes the format of the tokenizer.json file which will be supported in
#   onnxruntime-genai 0.5)
#

import argparse
import os
import time
import json
import shutil
import logging
from fnmatch import fnmatch
from queue import Queue
import subprocess
from packaging.version import Version
from huggingface_hub import snapshot_download
import onnxruntime_genai as og
import onnxruntime_genai.models.builder as model_builder
from transformers import AutoTokenizer
from lemonade.state import State
from lemonade.tools import FirstTool
import lemonade.common.status as status
import lemonade.common.printing as printing
from lemonade.tools.huggingface_load import get_base_model, is_offline
from lemonade.tools.adapter import (
    ModelAdapter,
    TokenizerAdapter,
    PassthroughTokenizerResult,
)
from lemonade.cache import Keys
from lemonade_install.install import (
    get_ryzen_ai_version_info,
    get_oga_npu_dir,
    get_oga_hybrid_dir,
    SUPPORTED_RYZEN_AI_SERIES,
)


# ONNX Runtime GenAI models will be cached in this subfolder of the lemonade cache folder
oga_models_path = "oga_models"

# ONNX Runtime GenAI model builder tool uses this subfolder of the lemonade cache as its cache
oga_model_builder_cache_path = "model_builder"

# Mapping from processor to execution provider, used in pathnames and by model_builder
execution_providers = {
    "cpu": "cpu",
    "npu": "npu",
    "igpu": "dml",
    "hybrid": "hybrid",
    "cuda": "cuda",
}


class OrtGenaiTokenizer(TokenizerAdapter):
    def __init__(self, model: og.Model, hf_tokenizer: AutoTokenizer):
        super().__init__(hf_tokenizer)
        # Initialize OGA tokenizer
        self.tokenizer = og.Tokenizer(model)

        # Placeholder value since some code will try to query it
        # If we actually need this to return a proper value, then
        # og.GeneratorParams.eos_token_id has it
        self.eos_token_id = None

    def __call__(self, prompt: str, return_tensors="np"):
        tokens = self.tokenizer.encode(prompt)
        return PassthroughTokenizerResult(tokens)

    # pylint: disable=unused-argument
    def decode(self, response, skip_special_tokens=True) -> str:
        return self.tokenizer.decode(response)


class OrtGenaiStreamer:
    def __init__(self, tokenizer: OrtGenaiTokenizer, timeout=None):
        self.tokenizer = tokenizer
        self.text_queue = Queue()
        self.stop_signal = None
        self.timeout = timeout

    def add_text(self, text: str):
        self.text_queue.put(text, timeout=self.timeout)

    def done(self):
        self.text_queue.put(self.stop_signal, timeout=self.timeout)

    def __iter__(self):
        return self

    def __next__(self):
        value = self.text_queue.get(timeout=self.timeout)
        if value == self.stop_signal:
            raise StopIteration()
        else:
            return value


class OrtGenaiModel(ModelAdapter):

    def __init__(self, input_folder):
        super().__init__()
        self.model = og.Model(input_folder)
        self.type = "ort-genai"
        self.config = self.load_config(input_folder)

    def load_config(self, input_folder):
        rai_config_path = os.path.join(input_folder, "rai_config.json")
        if os.path.exists(rai_config_path):
            with open(rai_config_path, "r", encoding="utf-8") as f:
                max_prompt_length = json.load(f)["max_prompt_length"]["1.4.1"]
        else:
            max_prompt_length = None

        config_path = os.path.join(input_folder, "genai_config.json")
        if os.path.exists(config_path):
            with open(config_path, "r", encoding="utf-8") as f:
                config_dict = json.load(f)
                if max_prompt_length:
                    config_dict["max_prompt_length"] = max_prompt_length
                return config_dict
        return None

    def generate(
        self,
        input_ids,
        max_new_tokens=512,
        min_new_tokens=0,
        do_sample=True,
        top_k=50,
        top_p=1.0,
        temperature=0.7,
        streamer: OrtGenaiStreamer = None,
        pad_token_id=None,
        stopping_criteria=None,
        max_length=None,
        random_seed=1,
    ):
        params = og.GeneratorParams(self.model)

        prompt_length = len(input_ids)
        max_prompt_length = self.config.get("max_prompt_length")
        if max_prompt_length and prompt_length > max_prompt_length:
            raise ValueError(
                f"This prompt (length {prompt_length}) exceeds the model's "
                f"maximum allowed prompt length ({max_prompt_length})."
            )

        # There is a breaking API change in OGA 0.6.0
        # Determine whether we should use the old or new APIs
        # This also supports 0.6.0.dev0, which evaluates to less than 0.6.0 in Version
        use_oga_post_6_api = (
            Version(og.__version__) >= Version("0.6.0") or "0.6.0" in og.__version__
        )
        use_oga_pre_6_api = not use_oga_post_6_api

        if pad_token_id:
            params.pad_token_id = pad_token_id

        # Handle max_length and max_new_tokens
        if max_length and max_new_tokens:
            logging.warning(
                "Both max_length and max_new_tokens were provided. "
                "max_length will take precedence. "
                "When setting max_length, please explicitly set max_new_tokens to None."
            )
        max_length_to_use = None
        if max_length:
            max_length_to_use = max_length
        elif max_new_tokens:
            max_length_to_use = prompt_length + max_new_tokens

        min_length = prompt_length + min_new_tokens

        if use_oga_pre_6_api:
            params.input_ids = input_ids

        if random_seed is None:
            random_seed = -1  # In og.Generator, -1 = seed with random device

        if self.config and "search" in self.config:
            search_config = self.config["search"]
            params.set_search_options(
                do_sample=search_config.get("do_sample", do_sample),
                top_k=search_config.get("top_k", top_k),
                top_p=search_config.get("top_p", top_p),
                temperature=search_config.get("temperature", temperature),
                max_length=max_length_to_use,
                min_length=min_length,
                early_stopping=search_config.get("early_stopping", False),
                length_penalty=search_config.get("length_penalty", 1.0),
                num_beams=search_config.get("num_beams", 1),
                num_return_sequences=search_config.get("num_return_sequences", 1),
                repetition_penalty=search_config.get("repetition_penalty", 1.0),
                past_present_share_buffer=search_config.get(
                    "past_present_share_buffer", True
                ),
                random_seed=random_seed,
                # Not currently supported by OGA
                # diversity_penalty=search_config.get('diversity_penalty', 0.0),
                # no_repeat_ngram_size=search_config.get('no_repeat_ngram_size', 0),
            )
        else:
            params.set_search_options(
                do_sample=do_sample,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                max_length=max_length_to_use,
                min_length=min_length,
                random_seed=random_seed,
            )
        params.try_graph_capture_with_max_batch_size(1)

        generator = og.Generator(self.model, params)

        if streamer is None:
            prompt_start_time = time.perf_counter()
            if use_oga_post_6_api:
                generator.append_tokens(input_ids)
            if use_oga_pre_6_api:
                generator.compute_logits()
            generator.generate_next_token()
            prompt_end_time = time.perf_counter()

            self.time_to_first_token = prompt_end_time - prompt_start_time

            if max_new_tokens > 1:

                token_gen_times = []
                while not generator.is_done():
                    token_gen_start_time = time.perf_counter()
                    if use_oga_pre_6_api:
                        generator.compute_logits()
                    generator.generate_next_token()
                    token_gen_end_time = time.perf_counter()

                    token_gen_times.append(token_gen_end_time - token_gen_start_time)

                if token_gen_times:
                    # List will be empty if we generated 1 or 0 tokens, and we don't
                    # want a divide-by-zero error in those cases
                    avg_token_gen_latency_s = sum(token_gen_times) / len(
                        token_gen_times
                    )
                    self.tokens_per_second = 1 / avg_token_gen_latency_s

            return [generator.get_sequence(0)]
        else:
            if use_oga_post_6_api:
                generator.append_tokens(input_ids)
            tokenizer_stream = streamer.tokenizer.tokenizer.create_stream()

            stop_early = False

            while not generator.is_done() and not stop_early:
                if use_oga_pre_6_api:
                    generator.compute_logits()
                generator.generate_next_token()

                new_token = generator.get_next_tokens()[0]
                new_text = tokenizer_stream.decode(new_token)

                streamer.add_text(new_text)

                if stopping_criteria is not None:
                    if stopping_criteria[0].stop_event.is_set():
                        stop_early = True

            streamer.done()

    def _model_call(self, input_ids):
        """
        Run the model on input_ids and get logits.

        This method directly accesses model logits rather than using the full generate pipeline for
        several important reasons:
        1. Purpose: We need raw logits from a single forward pass, while generate() is optimized for
           producing multiple tokens through iterative inference
        2. Efficiency: Direct access is more efficient for logprob calculations with no
           sampling overhead
        3. Precision: Logprob calculations require exact control over input-to-output mapping
        4. Consistency: Similar approach used in both HF and OGA implementations

        Args:
            input_ids: Input token IDs

        Returns:
            Logits for each token in the sequence
        """
        import torch

        # Setup generator params
        params = og.GeneratorParams(self.model)

        # Configure for a simple forward pass
        params.set_search_options(
            do_sample=False,
            temperature=0.0,
            max_length=len(input_ids),
        )

        # Initialize generator
        generator = og.Generator(self.model, params)

        # Feed tokens to model based on API version
        generator.append_tokens(input_ids)

        # Extract logits - this returns a list of logits tensors
        logits = generator.get_output("logits")

        # Convert to torch tensor for easier processing
        return torch.tensor(logits[0])

    def _select_cont_toks(self, logits, context_len, continuation_tokens):
        """
        Select and process logits for continuation tokens.

        Args:
            logits: Full sequence logits
            context_len: Length of context tokens
            continuation_tokens: List or tensor of continuation token IDs

        Returns:
            Log probabilities for continuation tokens
        """
        import torch

        # Extract relevant logits for continuation prediction (shift by one)
        cont_logits = logits[
            context_len - 1 : context_len - 1 + len(continuation_tokens)
        ]

        # Convert to torch tensors if needed
        if not isinstance(continuation_tokens, torch.Tensor):
            continuation_tokens = torch.tensor(continuation_tokens, dtype=torch.long)

        # Apply log softmax to get log probabilities
        log_probs = torch.log_softmax(cont_logits, dim=-1)

        # Get log probs for the specific continuation tokens
        token_log_probs = torch.gather(
            log_probs, 1, continuation_tokens.unsqueeze(-1)
        ).squeeze(-1)

        return token_log_probs

    def compute_logprobs(
        self, text, tokenizer, prompt_length=None, logprobs=None, echo=False
    ):
        """
        Compute log probabilities for all tokens in the given text.

        Args:
            text: The full text to analyze (e.g., prompt + completion)
            prompt_length: Number of tokens in the prompt. If provided and echo=False,
                only completion tokens after this position will be returned.
            logprobs: If not None, return log probabilities. Value indicates how many top
                alternatives to return. If True but not an integer, defaults to 5 alternatives.
            echo: If True, include logprobs for prompt tokens. If False, only return logprobs
                for completion tokens.

        Returns:
            - text_offset: Character offsets for each token in the text
            - token_logprobs: Log probability for each token
            - tokens: The actual tokens used
            - top_logprobs: Top alternative log probabilities for each position
        """
        import torch

        if tokenizer is None:
            raise ValueError("Tokenizer is required for logprob calculation")

        # Encode the full text
        tokens = tokenizer(text).input_ids  # pylint: disable=E1102

        # Track character offsets for each token
        text_offset = []
        start_idx = 0

        token_strings = []
        for token_id in tokens:
            token_str = tokenizer.decode([token_id])
            token_strings.append(token_str)

            # Calculate character offsets for tokens - handles cases where tokens
            # may not directly match in the original text due to encoding differences,
            # special characters, or tokenization artifacts
            try:
                pos = text[start_idx:].find(token_str)
                if pos != -1:
                    text_offset.append(start_idx + pos)
                    start_idx += pos + len(token_str)
                else:
                    text_offset.append(start_idx)
            except (TypeError, ValueError, UnicodeError):
                # Fallback to current position when matching fails due to encoding issues
                text_offset.append(start_idx)

        # Get logits from model
        logits = self._model_call(tokens)

        # Calculate log probabilities for each token
        all_log_probs = torch.log_softmax(logits, dim=-1)

        # The first token doesn't have a conditional probability
        # For tokens after the first, get the predicted probability
        token_log_probs = []
        top_logprobs_list = []

        # For each position, get the actual token probability and top alternatives
        for i in range(len(tokens)):
            # Get previous token position logits
            if i > 0:  # First token has no preceding context
                prev_logits = all_log_probs[i - 1]
                curr_token_id = tokens[i]
                # Get probability of the actual token that appeared
                token_logprob = prev_logits[curr_token_id].item()
                token_log_probs.append(token_logprob)

                # Get top-k alternatives if requested
                if logprobs is not None:
                    num_alternatives = logprobs if isinstance(logprobs, int) else 5
                    topk_values, topk_indices = torch.topk(
                        prev_logits, min(num_alternatives, prev_logits.size(-1))
                    )

                    # Create dictionary of token: logprob
                    position_logprobs = {}
                    for val, idx in zip(topk_values.tolist(), topk_indices.tolist()):
                        token_str = tokenizer.decode([idx])
                        position_logprobs[token_str] = val

                    top_logprobs_list.append(position_logprobs)
            else:
                # For the first token, we don't have a conditional probability
                token_log_probs.append(None)
                top_logprobs_list.append({})

        # If we don't want to echo prompt tokens, filter them out
        if not echo and prompt_length is not None:
            # Ensure prompt_length is within bounds
            prompt_length = min(prompt_length, len(tokens))

            # Filter results to only include completion tokens
            if prompt_length < len(tokens):
                filtered_text_offset = text_offset[prompt_length:]
                filtered_token_logprobs = token_log_probs[prompt_length:]
                filtered_tokens = token_strings[prompt_length:]
                filtered_top_logprobs = top_logprobs_list[prompt_length:]

                return (
                    filtered_text_offset,
                    filtered_token_logprobs,
                    filtered_tokens,
                    filtered_top_logprobs,
                )
            else:
                # No completion tokens
                return [], [], [], []

        return text_offset, token_log_probs, token_strings, top_logprobs_list


class OgaLoad(FirstTool):
    """
    Tool that loads an LLM in OnnxRuntime-GenAI for use with CPU or DirectML execution providers.

    Input: path to a checkpoint.
        Supported choices for cpu and igpu from HF model repository:
            LLM models on Huggingface supported by model_builder.  See documentation
            (https://github.com/lemonade-sdk/lemonade/blob/main/docs/ort_genai_igpu.md)
            for supported models.
        Supported choices for npu from HF model repository:
            Models on Hugging Face that follow the "amd/**-onnx-ryzen-strix" pattern
        Local models for cpu, igpu, or npu:
            The specified checkpoint is converted to a local path, via mapping to lower case
            and replacing '/' with '_'.  If this model already exists in the 'models' folder
            of the lemonade cache and if it has a subfolder <device>-<dtype>, then this model
            will be used.  If the --force flag is used and the model is built with model_builder,
            then it will be rebuilt.



    Output:
        state.model: handle to a Huggingface-style LLM loaded on DirectML device
        state.tokenizer = Huggingface-style LLM tokenizer instance
        state.dtype = data type of the model on DirectML device
        state.checkpoint = name of the checkpoint used to load state.model

    Note: This tool expects the onnxruntime-genai-directml library to be pre-installed.
            If that library is not installed, this tool will not load.
    """

    unique_name = "oga-load"

    def __init__(self):
        super().__init__(monitor_message="Loading OnnxRuntime-GenAI model")

        self.status_stats = [
            Keys.DTYPE,
            Keys.DEVICE,
            Keys.LOCAL_MODEL_FOLDER,
        ]

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Load model in onnxruntime-genai (OGA)",
            add_help=add_help,
        )

        parser.add_argument(
            "-ip",
            "--input_path",
            default="",
            help="the local huggingface model in your disk",
        )

        parser.add_argument(
            "-d",
            "--device",
            choices=["igpu", "npu", "cpu", "hybrid", "cuda"],
            default="igpu",
            help="Which device to load the model on to (default: igpu)",
        )

        parser.add_argument(
            "--dtype",
            choices=["int4", "fp16", "fp32"],
            required=True,
            help="Data type to load the model in",
        )

        parser.add_argument(
            "--int4-block-size",
            default=None,
            help="Specify the block_size for int4 quantization.",
            choices=[16, 32, 64, 128, 256],
            type=int,
        )

        parser.add_argument(
            "--force",
            action="store_true",
            help="Forces downloading of Hugging-Face model again (if changed).  Additionally for"
            " cpu and igpu devices only, forces model_builder to run again on the HF model"
            " (changed or not).",
        )

        parser.add_argument(
            "--download-only",
            action="store_true",
            help="Download the model if needed, but don't load it",
        )

        parser.add_argument(
            "--trust-remote-code",
            action="store_true",
            help="Set this flag to use models whose code is on the Hugging Face hub rather "
            "than natively in the OnnxRuntime Gen AI libraries.  Please review the model code "
            "in advance as this is a security risk.",
        )

        parser.add_argument(
            "--subfolder",
            default=None,
            help="Subfolder where model is located <LEMONADE CACHE>/oga_models/<MODELNAME>"
            "/<SUBFOLDER>, default is <EP for device>-<dtype>.  The EPs are: "
            f'{", ".join([value + " for " + key for key, value in execution_providers.items()])}.',
        )

        return parser

    @staticmethod
    def _validate_model_configuration(device, dtype, checkpoint):
        """
        Validate if the device, dtype, platform and checkpoint combination are consistent with
        HuggingFace checkpoint naming conventions and specifically for AMD models for NPU
        and hybrid flows.

        Returns True if device, dtype, and model are consistent.
        """

        hf_supported_models = {
            "cpu": {"int4": "*/*", "fp32": "*/*"},
            "igpu": {"int4": "*/*", "fp16": "*/*"},
            "npu": {"int4": "*/*"},
            "hybrid": {"int4": "*/*"},
            "cuda": {"int4": "*/*", "fp16": "*/*"},
        }

        hf_supported = (
            device in hf_supported_models
            and dtype in hf_supported_models[device]
            and fnmatch(checkpoint, hf_supported_models[device][dtype])
        )
        return hf_supported

    @staticmethod
    def _setup_model_paths(
        state, checkpoint, device, dtype, subfolder, int4_block_size
    ):
        """
        Determines and returns the following model path information for models produced by OGA
        model builder:

           full_model_path - Full path to where the OGA model files are stored.
           oga_models_subfolder - The subfolder of the oga_models folder where the model files
                are stored.  (<full_model_path> = <oga_models>/<oga_models_subfolder>)
                This subfolder is usually
                  <checkpoint_string>/<device>-<dtype>[-block-<int4_block_size]>
                but the if the argument subfolder is not None it will override the latter part
                of this path.
           model_exists_locally - True if full_model_path is a folder that contains files

        Note: Model files already in ONNX format on Hugging Face will be stored in the
            Hugging Face cache, not this folder.  The <oga_models> folder contains model
            files that have locally been quantized/converted to OGA format and any other
            models that have been manually added by the user.
        """
        if subfolder is None:
            subfolder = f"{execution_providers[device]}-{dtype}"
            subfolder += (
                f"-block-{int4_block_size}"
                if dtype == "int4" and int4_block_size is not None
                else ""
            )

        # First, check in the lemonade oga_models cache
        oga_models_subfolder = os.path.join(
            checkpoint.replace("/", "_").lower(), subfolder
        )
        full_model_path = os.path.join(
            state.cache_dir, oga_models_path, oga_models_subfolder
        )
        model_exists_locally = os.path.isdir(full_model_path) and os.listdir(
            full_model_path
        )

        # If not found in lemonade cache, check in Hugging Face cache
        if not model_exists_locally:
            try:
                snapshot_path = snapshot_download(
                    repo_id=checkpoint,
                    local_files_only=True,
                )

                # Check if the snapshot contains ONNX files
                if os.path.isdir(snapshot_path) and os.listdir(snapshot_path):
                    is_onnx_model = any(
                        filename.endswith(".onnx")
                        for filename in os.listdir(snapshot_path)
                    )

                    if is_onnx_model:
                        # If the model is in HF cache and has ONNX files, use it
                        full_model_path = snapshot_path
                        model_exists_locally = True
                        printing.log_info(
                            f"Found ONNX model in Hugging Face cache: {full_model_path}"
                        )
            except Exception as e:  # pylint: disable=broad-exception-caught
                # Log any errors but continue with the original path
                printing.log_info(f"Error checking Hugging Face cache: {e}")

        return full_model_path, model_exists_locally

    @staticmethod
    def _update_hybrid_custom_ops_library_path(full_model_path):
        """
        Modifies the genai_config.json file in the hybrid model folder to set the custom_ops_library
        path to the location of the onnx_custom_ops.dll in the current environment.
        This is needed for hybrid inference.
        """
        oga_path, version = get_oga_hybrid_dir()

        if "1.3.0" in version:
            custom_ops_path = os.path.join(
                oga_path,
                "onnx_utils",
                "bin",
                "onnx_custom_ops.dll",
            )
        else:
            custom_ops_path = os.path.join(oga_path, "libs", "onnx_custom_ops.dll")

        # Insert the custom_ops_path into the model config file
        config_path = os.path.join(full_model_path, "genai_config.json")
        if os.path.exists(config_path):
            with open(config_path, "r", encoding="utf-8") as f:
                config = json.load(f)

            if (
                "model" in config
                and "decoder" in config["model"]
                and "session_options" in config["model"]["decoder"]
            ):
                config["model"]["decoder"]["session_options"][
                    "custom_ops_library"
                ] = custom_ops_path

            with open(config_path, "w", encoding="utf-8") as f:
                json.dump(config, f, indent=4)

        else:
            printing.log_info(
                f"Model's `genai_config.json` not found in {full_model_path}"
            )

    @staticmethod
    def _is_preoptimized_model(input_model_path):
        """
        Checks if the 'custom_ops_library' field exists in the genai_config.json file
        to determine if this is a pre-optimized model for hybrid as well
        as NPU only.

        Args:
            input_model_path (str): Path to the input model directory.

        Returns:
            bool: True if 'custom_ops_library' exists, False otherwise.
        """
        config_path = os.path.join(input_model_path, "genai_config.json")
        if not os.path.exists(config_path):
            printing.log_info(f"Model's `genai_config.json` not found in {config_path}")
            return False

        with open(config_path, "r", encoding="utf-8") as f:
            config = json.load(f)
        if (
            "model" in config
            and "decoder" in config["model"]
            and "session_options" in config["model"]["decoder"]
        ):
            return "custom_ops_library" in config["model"]["decoder"]["session_options"]
        return False

    @staticmethod
    def _download_and_build_safetensors_model(
        checkpoint, device, dtype, full_model_path, int4_block_size, input_path, state
    ):
        """
        Uses OGA model builder to quantize safetensors format model and convert to ONNX
        format.  The model files are saved to the full_model_path folder.
        """
        printing.log_info(f"Building {checkpoint} for {device} using {dtype}")
        extra_options = {}
        if int4_block_size is not None:
            extra_options["int4-block-size"] = int4_block_size
        try:
            model_builder.create_model(
                checkpoint,
                input_path,
                full_model_path,
                dtype,
                execution_providers[device],
                os.path.join(state.cache_dir, oga_model_builder_cache_path),
                **extra_options,
            )
        except NotImplementedError as e:
            raise NotImplementedError("[Model builder] " + str(e)) from e
        except OSError as e:
            raise ValueError("[Model builder] " + str(e)) from e

        return full_model_path

    @staticmethod
    def _setup_npu_environment():
        """
        Sets up environment for NPU flow of ONNX model and returns saved state to be restored
        later in cleanup.
        """
        oga_path, version = get_oga_npu_dir()

        if not os.path.exists(os.path.join(oga_path, "libs", "onnxruntime.dll")):
            raise RuntimeError(
                f"Cannot find libs/onnxruntime.dll in lib folder: {oga_path}"
            )

        # Save current state so they can be restored after inference.
        saved_state = {"cwd": os.getcwd(), "path": os.environ["PATH"]}

        # Setup NPU environment (cwd and path will be restored later)
        os.chdir(oga_path)
        os.environ["PATH"] = (
            os.path.join(oga_path, "libs") + os.pathsep + os.environ["PATH"]
        )
        if "1.3.0" in version:
            os.environ["DD_ROOT"] = ".\\bins"
            os.environ["DEVICE"] = "stx"
            os.environ["XLNX_ENABLE_CACHE"] = "0"

        return saved_state

    @staticmethod
    def _setup_hybrid_environment():
        """
        Sets up the environment for the Hybrid flow and returns saved state to be restored later
        in cleanup.
        """
        # Determine the Ryzen AI OGA version and hybrid artifacts path
        oga_path, version = get_oga_hybrid_dir()

        if "1.3.0" in version:
            dst_dll = os.path.join(
                oga_path,
                "onnx_utils",
                "bin",
                "DirectML.dll",
            )
            if not os.path.isfile(dst_dll):
                # Artifacts 1.3.0 has DirectML.dll in different subfolder, so copy it to the
                # correct place.  This should not be needed in later RAI release artifacts.
                src_dll = os.path.join(
                    oga_path,
                    "onnxruntime_genai",
                    "lib",
                    "DirectML.dll",
                )
                os.makedirs(os.path.dirname(dst_dll), exist_ok=True)
                shutil.copy2(src_dll, dst_dll)

        saved_state = None
        return saved_state

    @staticmethod
    def _load_model_and_setup_state(
        state, full_model_path, checkpoint, trust_remote_code
    ):
        """
        Loads the OGA model from local folder and then loads the tokenizer.
        Will auto-detect if we're offline.
        """
        try:
            state.model = OrtGenaiModel(full_model_path)
        except Exception as e:
            if "invalid unordered_map<K, T>" in str(e):
                raise ValueError(
                    "Error initializing model: Invalid configuration detected.\n"
                    "Please check the following:\n"
                    f"1. Please check your model's config file in {full_model_path} "
                    "and ensure custom_ops_library points to the valid "
                    "onnx_custom_ops.dll path.\n"
                    "2. Make sure the NPU driver is loaded.\n"
                    "3. Make sure hybrid has been installed on a Ryzen AI "
                    f"{'or '.join(SUPPORTED_RYZEN_AI_SERIES)}-series processor."
                ) from e
            raise

        # Auto-detect offline mode
        offline = is_offline()

        try:
            # Always try to use local files first
            local_files_only = True

            hf_tokenizer = AutoTokenizer.from_pretrained(
                full_model_path,
                local_files_only=local_files_only,
                trust_remote_code=trust_remote_code,
            )
        except ValueError as e:
            if "trust_remote_code" in str(e):
                raise ValueError(
                    "This model requires you to execute code from the repo.  Please review it "
                    "and if you trust it, then use the `--trust-remote-code` flag with oga-load."
                )

            if offline and "Can't load tokenizer for" in str(e):
                raise ValueError(
                    f"Cannot load tokenizer for {checkpoint} in offline mode. "
                    f"The tokenizer files may not be available locally in {full_model_path}."
                )
            raise

        state.tokenizer = OrtGenaiTokenizer(
            state.model.model,
            hf_tokenizer,
        )

        status.add_to_state(state=state, name=checkpoint, model=checkpoint)

    @staticmethod
    def _cleanup_environment(saved_state):
        """
        Restores environment to its original state after inference is complete.
        """
        if saved_state:
            os.chdir(saved_state["cwd"])
            os.environ["PATH"] = saved_state["path"]

    def _generate_model_for_hybrid_or_npu(
        self, output_model_path, device, input_model_path
    ):
        """
        Uses a subprocess to run the 'model_generate' command for hybrid or npu devices.
        """

        # Determine the appropriate flag based on the device type
        if device == "hybrid":
            device_flag = "--hybrid"
        elif device == "npu":
            device_flag = "--npu"
        else:
            raise ValueError(f"Unsupported device type for model generation: {device}")

        command = [
            "model_generate",
            device_flag,
            output_model_path,  # Output model directory
            input_model_path,  # Input model directory
        ]

        printing.log_info(f"Running command: {' '.join(command)}")
        try:
            with open(self.logfile_path, "w", encoding="utf-8") as log_file:
                subprocess.run(
                    command, check=True, text=True, stdout=log_file, stderr=log_file
                )
        except FileNotFoundError as e:
            error_message = (
                "The 'model_generate' package is missing from your system. "
                "Ensure all required packages are installed. "
                "To install it, run the following command:\n\n"
                "    lemonade-install --ryzenai <target> --build-model\n"
            )
            raise RuntimeError(error_message) from e

    def run(
        self,
        state: State,
        input: str,
        input_path: str = "",
        device: str = "igpu",
        dtype: str = "int4",
        int4_block_size: int = None,
        force: bool = False,
        download_only: bool = False,
        trust_remote_code=False,
        subfolder: str = None,
    ) -> State:
        # Auto-detect offline status
        offline = is_offline()
        if offline:
            printing.log_warning(
                "Network connectivity to huggingface.co not detected. Running in offline mode."
            )

        state.device = device
        state.dtype = dtype

        # Log initial stats
        state.save_stat(Keys.DTYPE, dtype)
        state.save_stat(Keys.DEVICE, device)
        if device in ["hybrid", "npu"]:
            ryzen_ai_version_info = get_ryzen_ai_version_info()
            state.save_stat(Keys.RYZEN_AI_VERSION_INFO, ryzen_ai_version_info)

        # Check if input is a local folder
        if os.path.isdir(input):
            # input is a local folder
            full_model_path = os.path.abspath(input)
            checkpoint = "local_model"
            state.checkpoint = checkpoint
            state.save_stat(Keys.CHECKPOINT, checkpoint)
            state.save_stat(Keys.LOCAL_MODEL_FOLDER, full_model_path)
            # See if there is a file ending in ".onnx" in this folder
            dir = os.listdir(input)
            has_onnx_file = any([filename.endswith(".onnx") for filename in dir])
            if not has_onnx_file:
                raise ValueError(
                    f"The folder {input} does not contain an ONNX model file."
                )
            if force:
                raise ValueError(
                    "Your input (-i, --input) points to a local folder, which is not "
                    "compatible with the force argument."
                )

        else:
            # input is a model checkpoint
            checkpoint = input
            state.checkpoint = checkpoint
            state.save_stat(Keys.CHECKPOINT, checkpoint)

            # Get base model information
            if not offline:
                base_model = get_base_model(checkpoint)
                if base_model is not None:
                    state.save_stat("base_model", base_model)

            # Setup paths
            full_model_path, model_exists_locally = self._setup_model_paths(
                state, checkpoint, device, dtype, subfolder, int4_block_size
            )

            # If in offline mode, we can only use locally available models
            if offline and not model_exists_locally:
                raise ValueError(
                    f"Model {checkpoint} is not available locally for {device} with {dtype}. "
                    f"Cannot download in offline mode. Check {full_model_path}"
                )

            # Handle download/build if needed
            if (not model_exists_locally) or force:
                if offline:
                    raise ValueError(
                        f"Cannot download or build model {checkpoint} in offline mode"
                    )

                # Validate configuration
                hf_supported = self._validate_model_configuration(
                    device, dtype, checkpoint
                )

                if not hf_supported:
                    raise ValueError(
                        "The (device, dtype, checkpoint) combination is not supported: "
                        f"({device}, {dtype}, {checkpoint})"
                    )
                input_model_path = snapshot_download(
                    repo_id=checkpoint,
                    ignore_patterns=["*.md", "*.txt"],
                    local_files_only=offline,
                )
                # Check if model is ONNX or safetensors
                is_onnx_model = any(
                    [
                        filename.endswith(".onnx")
                        for filename in os.listdir(input_model_path)
                    ]
                )
                is_preoptimized_onnx = is_onnx_model and self._is_preoptimized_model(
                    input_model_path
                )
                is_safetensors_model = any(
                    [
                        filename.endswith(".safetensors")
                        for filename in os.listdir(input_model_path)
                    ]
                )
                if not (is_onnx_model or is_safetensors_model):
                    raise ValueError(
                        f"The model {checkpoint} is not supported. "
                        "It does not contain ONNX or safetensors files."
                    )
                if device in ["npu", "hybrid"]:
                    if is_onnx_model:
                        if is_preoptimized_onnx:
                            # Use HuggingFace cache path as it is
                            full_model_path = input_model_path
                        else:
                            # If ONNX but not modified yet for Hybrid or NPU,
                            # needs further optimization
                            self._generate_model_for_hybrid_or_npu(
                                full_model_path,
                                device,
                                input_model_path,
                            )
                    elif is_safetensors_model:
                        config_path = os.path.join(input_model_path, "config.json")
                        if os.path.exists(config_path):
                            with open(config_path, "r", encoding="utf-8") as f:
                                config = json.load(f)
                            if "quantization_config" in config:
                                # If quantized, use subprocess to generate the model
                                self._generate_model_for_hybrid_or_npu(
                                    full_model_path, device, input_model_path
                                )
                            else:
                                raise ValueError(
                                    f"The safetensors model {checkpoint} is not quantized. "
                                    "Only quantized safetensors models are supported"
                                    " on npu or hybrid targets."
                                )
                        else:
                            raise ValueError(
                                f"config.json not found for safetensors model: {checkpoint}"
                            )
                    else:
                        raise ValueError(
                            f"Unsupported model type for checkpoint: {checkpoint}"
                        )
                else:
                    if is_onnx_model:
                        # Use HuggingFace cache path as it is
                        full_model_path = input_model_path
                    else:
                        self._download_and_build_safetensors_model(
                            checkpoint,
                            device,
                            dtype,
                            full_model_path,
                            int4_block_size,
                            input_path,
                            state,
                        )
                    state.save_stat(Keys.LOCAL_MODEL_FOLDER, full_model_path)

        # Load model if download-only argument is not set
        if not download_only:

            saved_env_state = None
            try:
                if device == "npu":
                    saved_env_state = self._setup_npu_environment()
                    # Set USE_AIE_RoPE based on model type
                    os.environ["USE_AIE_RoPE"] = (
                        "0" if "phi-" in checkpoint.lower() else "1"
                    )
                elif device == "hybrid":
                    saved_env_state = self._setup_hybrid_environment()
                    self._update_hybrid_custom_ops_library_path(full_model_path)

                self._load_model_and_setup_state(
                    state, full_model_path, checkpoint, trust_remote_code
                )
            finally:
                self._cleanup_environment(saved_env_state)

        return state


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import os
import sys
from pathlib import Path

import torch
from transformers import AutoProcessor
from lemonade.state import State
from lemonade.tools import Tool
import lemonade.common.printing as printing
import lemonade.common.build as build
from lemonade_install.install import DEFAULT_QUARK_DIR


class QuarkQuantize(Tool):
    """
    Quantize a model using the Quark Quantization tool.

    This Tool performs the following steps:
    1. Downloads and extracts necessary resources from AMD Quark Web Page.
    2. Based on the target model, it prepares the model, tokenizer, and calibration data.
    3. Optionally quantizes, freezes, and exports the model.
    4. Optionally evaluates the model.

    Required Input State:
        - state.model.model: Pretrained model instance to be quantized.
        - state.tokenizer: Tokenizer instance from Hugging Face.
    Output:
        - Modifies `state` with quantized and optionally exported model.

    See docs/quark.md for more details.
    """

    unique_name = "quark-quantize"

    def __init__(self):
        super().__init__(monitor_message="Quark Quantizing model")

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Quantize a model using Quark",
            add_help=add_help,
        )
        parser.add_argument(
            "--device",
            default="cpu",
            choices=["cuda", "cpu"],
            help="Device for running the quantizer",
        )
        parser.add_argument("--multi-gpu", action="store_true")
        parser.add_argument(
            "--data-type",
            default="auto",
            choices=["auto", "float16", "bfloat16", "float32"],
            help="Input datatype of the model",
        )
        parser.add_argument(
            "--seq-len", type=int, default=512, help="Sequence length of data"
        )
        parser.add_argument(
            "--batch-size", type=int, default=1, help="Batch size for calibration."
        )
        parser.add_argument(
            "--num-fewshot",
            type=int,
            default=None,
            metavar="N",
            help="Number of examples in few-shot context",
        )
        parser.add_argument(
            "--output-dir", default=None, help="Output directory for exported model"
        )
        parser.add_argument(
            "--no-weight-matrix-merge",
            action="store_true",
            help="If set, merges onnx model and weight \
            together before export.\
            By default, for onnx export, spits out a model.onnx and a model.weights",
        )
        parser.add_argument(
            "--dataset",
            default="pileval",
            choices=[
                "pileval",
                "wikitext",
                "pileval_for_awq_benchmark",
                "wikitext_for_gptq_benchmark",
                "HuggingFaceH4/ultrachat_200k",
            ],
            help="Dataset for calibration",
        )
        parser.add_argument(
            "--num-calib-data",
            type=int,
            default=512,
            help="Number of samples for calibration.",
        )

        # See docs/quark.md for more details.
        parser.add_argument(
            "--quant-scheme",
            type=str,
            default=None,
            choices=[
                "w_fp8_a_fp8",
                "w_int4_per_channel_sym",
                "w_uint4_per_group_asym",
                "w_int4_per_group_sym",
                "w_uint4_a_bfloat16_per_group_asym",
                "w_int8_per_tensor_sym",
                "w_int8_per_group_sym",
                "w_uint8_per_group_asym",
                "w_int8_a_int8_per_tensor_sym",
                "w_int8_a_int8_per_tensor_sym_dynamic",
                "w_uint8_a_uint8_per_tensor_asym",
                "w_fp8_a_fp8_o_fp8",
                "w_mx_fp8",
                "w_mx_fp8_a_mx_fp8",
                "w_int8_a_int8_per_token_dynamic",
                "w_bfp16",
                "w_bfp16_a_bfp16",
                "w_mx6",
                "w_mx6_a_mx6",
                "w_fp8_per_channel_sym",
                "w_int4_per_channel_asym",
                "w_int4_per_group_asym",
                "w_uint4_per_group_sym",
                "w_uint4_per_channel_sym",
                "w_uint4_per_channel_asym",
                "w_int8_per_tensor_percentile",
                "w_int8_per_tensor_mse",
                "w_uint8_per_tensor_percentile",
                "w_uint8_per_tensor_mse",
                "w_mx_fp4_per_group_sym",
                "w_mx_fp6_e3m2_per_group_sym",
                "w_mx_fp6_e2m3_per_group_sym",
                "w_mx_int8_per_group_sym",
                "w_uint4_per_channel_a_int8_per_tensor",
                "w_uint4_per_group_a_int8_per_tensor",
                "w_bfp16_per_group_sym",
                None,
            ],
            help="Supported quantization schemes in Quark",
        )
        parser.add_argument(
            "--quant-algo",
            type=str,
            default=None,
            choices=["awq", "gptq", "autosmoothquant", None],
            help="Support quantization algorithms in Quark",
        )
        parser.add_argument(
            "--pre-optimization-config-file-path",
            type=str,
            default=None,
            help="The JSON file path of pre-optimization config",
        )
        parser.add_argument(
            "--quant-algo-config-file-path",
            type=str,
            default=None,
            help="The JSON file path of quantization algorithm config",
        )
        parser.add_argument(
            "--group-size",
            type=int,
            default=128,
            help="Group size for per_group quantization",
        )
        parser.add_argument(
            "--pack-method",
            type=str,
            default="reorder",
            choices=["order", "reorder"],
            help="Pack method for awq_export",
        )
        parser.add_argument(
            "--exclude-layers",
            type=str,
            nargs="*",
            default=None,
            help="List of layers to exclude from quantization.",
        )
        parser.add_argument(
            "--kv-cache-dtype",
            default=None,
            choices=["fp8", None],
            help="KV Cache dtype.",
        )
        parser.add_argument(
            "--pre-quantization-optimization",
            action="append",
            default=[],
            choices=["rotation", "smoothquant"],
            help="Pre Quantization Optimization.",
        )
        parser.add_argument(
            "--model-export",
            default=None,
            action="append",
            choices=[
                None,
                "onnx",
                "vllm_adopted_safetensors",
                "quark_safetensors",
                "gguf",
            ],
            help="Model export format",
        )
        parser.add_argument(
            "--custom-mode",
            default="quark",
            type=str,
            choices=["quark", "awq", "fp8"],
            help="Custom mode for export \
                This is especially relevant for npu/hybrid export",
        )
        parser.add_argument(
            "--torch-compile",
            action="store_true",
            help="Compile the quantized model using torch.compile",
        )
        parser.add_argument(
            "--params-save", action="store_true", help="Save model params"
        )
        parser.add_argument(
            "--save-dir",
            help="Directory to save model parameters as \
            safetensors or pth, in the case when --params_save is used.",
        )
        parser.add_argument(
            "--log-severity-level", type=int, default=3, help="DEBUG=1, INFO=2, ERROR=3"
        )
        parser.add_argument("--skip-quantization", action="store_true")

        return parser

    def run(self, state: State, **kwargs) -> State:
        """
        Executes the QuarkQuantize process.

        Args:
            state (State): The current state of the process, containing necessary
            information such as cache directory and build name.
            **kwargs: Additional keyword arguments that may include:
            - output_dir (str): Directory to save the output model.
            - safetensors_model_dir (str): Directory to save the safetensors model.
            - save_dir (str): Directory to save model parameters.
            - safetensors_path (str): Path to the safetensors model.
            - quant_algo (str): The quantization algorithm to use.
            - quant_algo_config_file_path (str): Path to the quantization algorithm
              configuration file.
            - model_dir (str): Directory of the model.
        Returns:
            State: The updated state after the quantization process.
        Raises:
            Exception: If an error occurs during the QuarkQuantize process
            and when installation path does not exist.
        """

        try:

            if os.path.isdir(DEFAULT_QUARK_DIR):
                quark_llm_path = os.path.join(
                    DEFAULT_QUARK_DIR, "examples", "torch", "language_modeling"
                )
                sys.path.extend([quark_llm_path])
            else:
                raise FileNotFoundError(
                    f"The directory {DEFAULT_QUARK_DIR} does not exist. \
                        Please check your installation."
                )
            model_build_path = os.path.join(
                build.output_dir(state.cache_dir, state.build_name)
            )
            model_export_path = os.path.join(
                model_build_path,
                "exported_model",
                kwargs.get("quant_scheme"),
                kwargs.get("quant_algo"),
            )
            # Set default paths only if current values are None
            if kwargs.get("model_dir") is None:
                kwargs["model_dir"] = model_build_path
            if kwargs.get("output_dir") is None:
                kwargs["output_dir"] = model_export_path
            if kwargs.get("save_dir") is None:
                kwargs["save_dir"] = os.path.join(model_export_path, "model_params")

            from llm_utils.model_preparation import get_model_type

            model_type = get_model_type(state.model.model)

            quant_algo = kwargs.get("quant_algo")
            kwargs["quant_algo_config_file_path"] = os.path.join(
                quark_llm_path,
                "llm_ptq",
                "models",
                model_type,
                f"{quant_algo}_config.json",
            )

            self._quantize(state, **kwargs)

        except Exception as e:
            printing.log_error(f"Error during the QuarkQuantize process: {e}")
            raise
        return state

    def _quantize(self, state: State, **kwargs) -> None:
        """
        Main quantization and export process.

        This method is responsible for:
        - Loading the model and tokenizer.
        - Preparing the calibration dataset.
        - Quantizing the model.
        - Optionally exporting, compiling, and evaluating the model.
        """

        model = state.model.model
        tokenizer = state.tokenizer

        # Importing quark utils after adding to sys.path
        from llm_utils.data_preparation import get_calib_dataloader
        from llm_utils.model_preparation import get_model_type
        from llm_ptq.configuration_preparation import get_config, get_export_config
        from quark.torch import ModelQuantizer, ModelExporter, save_params

        # 1. Load Model
        printing.log_info("Loading model ...")
        model_type = get_model_type(model)

        # [mllama specifics]
        if model_type == "mllama" and kwargs.get("model_export") is not None:
            processor = AutoProcessor.from_pretrained(kwargs.get("model_dir"))
            export_dir = Path(kwargs.get("output_dir"))
            export_dir.mkdir(parents=True, exist_ok=True)
            processor.save_pretrained(kwargs.get("output_dir"))

        # 2. Load dataset
        printing.log_info("Loading dataset ...")
        main_device = model.device if kwargs.get("multi_gpu") else kwargs.get("device")
        calib_dataloader = get_calib_dataloader(
            dataset_name=kwargs.get("dataset"),
            tokenizer=tokenizer,
            batch_size=1,
            num_calib_data=kwargs.get("num_calib_data"),
            seqlen=kwargs.get("seq_len"),
            device=main_device,
        )

        # 3. Quantize model
        if not kwargs.get("skip_quantization"):
            printing.log_info("Starting quantization process ...")
            args = argparse.Namespace(**kwargs)
            quant_config = get_config(args, model_type)
            quant_config.log_severity_level = kwargs.get("log_severity_level", 3)
            quantizer = ModelQuantizer(quant_config)
            model = quantizer.quantize_model(model, calib_dataloader)
            printing.log_info("Quantization completed.")

            if (
                kwargs.get("model_export") is not None
                or kwargs.get("params_save")
                or kwargs.get("torch_compile")
            ):
                printing.log_info("Freezing the quantized model ...")
                model = quantizer.freeze(model)

        # 4. Export model
        if kwargs.get("model_export") is not None:
            printing.log_info("Exporting the model ...")
            export_path = kwargs.get("output_dir")

            args = argparse.Namespace(**kwargs)
            export_config = get_export_config(args, model_type)
            exporter = ModelExporter(config=export_config, export_dir=export_path)
            if "quark_safetensors" in kwargs.get("model_export"):
                printing.log_info("Exporting quark native json and safetensors...")
                with torch.no_grad():
                    quant_config = get_config(args, model_type)
                    exporter.export_model_info(
                        model,
                        quant_config=quant_config,
                        tokenizer=tokenizer,
                        custom_mode=kwargs.get("custom_mode"),
                    )
            if "vllm_adopted_safetensors" in kwargs.get("model_export"):
                printing.log_info("Exporting vllm adopted json and safetensors...")
                with torch.inference_mode():
                    exporter.export_model_info(
                        model,
                        model_type=model_type,
                        model_dtype=state.dtype,
                        export_type="vllm-adopt",
                    )
            if "onnx" in kwargs.get("model_export"):
                printing.log_info("Exporting onnx graph...")
                with torch.inference_mode():
                    batch_iter = iter(calib_dataloader)
                    input_args = next(batch_iter)
                    if kwargs.get("quant_scheme") in [
                        "w_int4_per_channel_sym",
                        "w_uint4_per_group_asym",
                        "w_int4_per_group_sym",
                        "w_uint4_a_bfloat16_per_group_asym",
                    ]:
                        uint4_int4_flag = True
                    else:
                        uint4_int4_flag = False
                    exporter.export_onnx_model(
                        model, input_args, uint4_int4_flag=uint4_int4_flag
                    )
            if "gguf" in kwargs.get("model_export"):
                printing.log_info("Exporting gguf model...")
                with torch.inference_mode():
                    exporter.export_gguf_model(
                        model, kwargs.get("model_dir"), model_type
                    )

        # 6. [Optional] Compile model
        if kwargs.get("torch_compile"):
            printing.log_info("torch.compile...")
            model = torch.compile(model)

        # 7. Save model parameters
        if kwargs.get("params_save"):
            printing.log_info("Saving model parameters ...")
            save_params(model, model_type=model_type, export_dir=kwargs.get("save_dir"))

        state.model.model = model
        state.dtype = model.dtype
        printing.log_info("QuarkQuantize process completed.")


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD
